from __future__ import annotations

from importlib                     import import_module
from importlib.resources           import files

from prompt_toolkit.formatted_text import HTML, ANSI
from prompt_toolkit.shortcuts      import print_formatted_text
from ptpython.repl                 import PythonRepl
from ptpython.python_input         import PythonInput
from rich.markdown                 import Markdown

from frplib.env        import environment
from frplib.exceptions import FrplibException
from frplib.protocols  import Renderable


#
# Help System
#

def info(obj_or_topic='topics', pager=False) -> None:
    """Accesses and displays help on a variety of playground topics.

    """
    def no_help():
        print_formatted_text(HTML('<violet>I could not find any guidance on that topic. '
                                  'Try info() for a list of starting points.</violet>'))

    topic = []
    if not isinstance(obj_or_topic, str):
        if hasattr(obj_or_topic, '__info__'):
            topic = obj_or_topic.__info__.split('::')
        else:
            help(obj_or_topic)
            return
    else:
        # Look for main level topic or search below
        topic = obj_or_topic.split('::')

    if topic:
        top_level = files('frplib.data') / 'playground-help'

        topic_path = top_level
        found = True
        for dir in topic[:-1]:
            topic_path = topic_path / dir
            if not topic_path.is_dir():  # Will fail if does not exist also
                found = False
                break
        if found:
            topic_path = topic_path / f'{topic[-1]}.md'
            if not topic_path.is_file():
                found = False
            else:
                help_text = topic_path.read_text()
                # Rich behaving oddly here
                code_theme = 'monokai' if environment.dark_mode else 'slate'
                info_text = Markdown(help_text, code_theme=code_theme)
                if pager:
                    with environment.console.pager():
                        environment.console.print(info_text)
                else:
                    environment.console.print(info_text)
        if not found:
            pass  # Search for topic in manifest
            no_help()  # Do this if search comes up empty
    else:
        no_help()


#
# Import Environment
#

playground_imports: dict[str, list[str]] = {
    # When there are submodules loaded, order from parent down
    # Note that symbols and module names cannot be the same
    # as we are adding both to globals()
    # ATTN: Consider *not* adding the module names to globals
    'kinds': [
        'Kind', 'ConditionalKind',
        'kind', 'conditional_kind',
        'unfold', 'clean', 'fast_mixture_pow', 'bayes',
        'constant', 'uniform', 'either',
        'symmetric', 'linear', 'geometric',
        'weighted_by', 'weighted_as', 'weighted_pairs',
        'arbitrary', 'integers', 'evenly_spaced', 'bin',
        'without_replacement', 'subsets', 'permutations_of',
    ],
    'statistics': [
        'Statistic', 'Condition', 'MonoidalStatistic',
        'is_statistic', 'statistic', 'condition', 'scalar_statistic',
        'tuple_safe', 'infinity',
        'fork', 'chain', 'compose',
        'Id', 'Scalar', '__', 'Proj', '_x_',
        'Sum', 'Product', 'Count', 'Max', 'Min', 'Abs', 'SumSq',
        'Sqrt', 'Floor', 'Ceil', 'NormalCDF',
        'Exp', 'Log', 'Log2', 'Log10',
        'Sin', 'Cos', 'Tan', 'ATan2', 'Sinh', 'Cosh', 'Tanh',
        'Mean', 'StdDev', 'Variance',
        'Diff', 'Diffs', 'Permute',
        'Constantly', 'Fork', 'MFork', 'ForEach', 'IfThenElse',
        'And', 'Or', 'Not', 'Xor', 'top', 'bottom',
    ],
    'expectations': ['E', 'D_'],
    'frps': [
        'FRP', 'frp', 'conditional_frp', 'shuffle',
    ],
    'calculate': ['substitute', 'substitute_with', 'substitution'],
    'numeric': ['numeric_exp', 'numeric_ln', 'numeric_log10', 'numeric_log2',
                'numeric_abs', 'numeric_sqrt', 'numeric_floor', 'numeric_ceil'],
    'quantity': ['as_quantity', 'qvec'],
    'symbolic': ['gen_symbol', 'is_symbolic', 'symbol'],
    'utils': [
        'clone', 'compose', 'const', 'every', 'frequencies',
        'identity', 'index_of', 'irange', 'iterate', 'lmap',
        'values', 'dim', 'codim', 'size', 'show', 'some',
    ],
    'vec_tuples': [
        'VecTuple',
        'as_numeric_vec', 'as_scalar', 'as_vec_tuple',
        'is_vec_tuple', 'vec_tuple',
    ],
    'extras': [
        'components',
    ],
}

# ATTN: maybe don't load the modules into globals?

def import_playground(globals) -> None:
    modules = playground_imports.keys()
    for module_name in modules:
        module = import_module(f'frplib.{module_name}')
        globals[module_name] = module
        for symbol_name in playground_imports[module_name]:
            globals[symbol_name] = getattr(module, symbol_name)

def remove_playground(globals) -> None:
    modules = playground_imports.keys()
    for module_name in modules:
        for symbol_name in playground_imports[module_name]:
            if symbol_name in globals:
                del globals[symbol_name]
        if module_name in globals:
            del globals[module_name]


#
# REPL Definition
#

class PlaygroundRepl(PythonRepl):
    def show_result(self, result: object) -> None:
        """
        Show __repr__ for an `eval` result and print to output.
        """
        if isinstance(result, Renderable) and not isinstance(result, type):
            # Don't call for classes e.g., VecTuple as a class not instance
            # Holding off on pager for now
            try:
                environment.console.print(result.__frplib_repr__())
            except Exception as e:
                environment.console.print(f'Could not print result due to an error: {str(e)}')
        else:
            formatted_text_output = self._format_result_output(result)

            if self.enable_pager:
                self.print_paginated_formatted_text(formatted_text_output)
            else:
                self.print_formatted_text(formatted_text_output)

        self.app.output.flush()
        if self.insert_blank_line_after_output:
            self.app.output.write("\n")

    def _handle_exception(self, e: BaseException) -> None:
        output = self.app.output

        if isinstance(e, FrplibException):
            # ATTN
            print_formatted_text(
                ANSI(environment.console_str(str(e))),
                output=output
            )
        else:
            tokens = self._format_exception_output(e)

            print_formatted_text(
                tokens,
                style=self._current_style,
                style_transformation=self.style_transformation,
                include_default_pygments_style=False,
                output=output,
            )
        output.flush()

    def _handle_keyboard_interrupt(self, e: KeyboardInterrupt) -> None:
        output = self.app.output

        output.write("\rOperation Interrupted\n\n")
        output.flush()

    def _add_to_namespace(self) -> None:
        """
        Add ptpython built-ins to global namespace.
        """
        globals = self.get_globals()

        # Add a 'get_ptpython', similar to 'get_ipython'
        def get_playground() -> PythonInput:
            return self

        self._injected_globals = [
            "get_playground",
            "info",
            "_running_in_playground"
        ]

        globals["get_playground"] = get_playground
        globals["info"] = info
        globals["_running_in_playground"] = True
        environment.interactive_mode()
        import_playground(globals)

    def _remove_from_namespace(self) -> None:
        """
        Remove added symbols from the globals.
        """
        globals = self.get_globals()
        for symbol in self._injected_globals:
            del globals[symbol]
        remove_playground(globals)



#
# Random Points in a Circle Example Chapter 0, Section 4.3
# Points in a Circle, With Constraints Example Chapter 0, Section 5
#
from math              import floor
from typing            import Literal, Union

from frplib.frps       import FRP, frp
from frplib.kinds      import (conditional_kind,
                               constant, uniform, weighted_by)
from frplib.utils      import compose, dim, irange

# ATTN! Need to update for new interfaces, e.g., domain, codim
# This def is temporary along these lines
from frplib.vec_tuples import as_vec_tuple
from frplib.exceptions import MismatchedDomain

def value_set(*vals):
    if len(vals) == 1 and hasattr(vals[0], '__next__'):
        # We have been given a single iterator/generator, use it
        vals = vals[0]
    vs = set([as_vec_tuple(v) for v in vals])
    dims = set(map(dim, vs))
    if len(dims) != 1:
        raise MismatchedDomain(f'Value set {vals} have different dimensions.')
    return vs


#
# Helpers
#

def y_points(x: int, radius: Union[int, float] = 5) -> list[int]:
    r = abs(radius)
    r_lo = floor(r)
    return [y for y in irange(-r_lo, r_lo) if y * y <= r * r - x * x]

num_y_points = compose(len, y_points)   # len `after` y_points

def points_inside(radius: Union[int, float] = 5) -> list[tuple[int, int]]:
    "Returns list of integer points within circle of given radius >= 0"
    r = abs(radius)
    r_lo = floor(r)
    vals = list(irange(-r_lo, r_lo))
    return [(x, y) for x in vals for y in vals
            if x * x + y * y <= r * r]


#
# FRP Factories
#

def circle_points(
        radius: Union[int, float] = 5,
        method: Literal["mixture", "jointly"] = "mixture"
) -> FRP:
    "Returns FRP for uniform random point inside circle of a given radius."
    if method == "jointly":
        return frp(uniform(points_inside(radius)))

    # Construct as a mixture
    r = abs(radius)
    r_lo = floor(r)
    if r == 0:
        return frp(constant(0, 0))

    x_kind = weighted_by(-r_lo, -r_lo + 1, ..., r_lo, weight_by=num_y_points)

    @conditional_kind(domain=value_set(irange(-r_lo, r_lo)))   # ATTN: remove value_set
    def y_kind(x):
        return uniform(y_points(x))

    return frp(x_kind >> y_kind)



from __future__ import annotations

import sys

from collections       import defaultdict
from collections.abc   import Iterable, Hashable
from functools         import reduce
from typing            import Callable, Generator, TypeVar, Union
from typing_extensions import Any, TypeGuard

from frplib.env        import environment
from frplib.exceptions import ConstructionError, OperationError
from frplib.protocols  import Renderable

#
# Generic
#

A = TypeVar('A')

def identity(x: A) -> A:
    "Returns its argument."
    return x

def const(a: A) -> Callable[[Any], A]:
    "Returns a constant function that returns the given value."
    def const_fn(x: Any) -> A:
        return a
    return const_fn


#
# Kinds and FRPs and Such
#

def values(x, scalarize=False) -> set:
    """Returns the set of values of a kind.

    Parameters:
    ----------
      x - a kind, or any object with a .values property
      scalarize [False] - if True, convert numeric scalars to floats

    Returns a set of possible values, with scalars unwrapped from
    their tuples.

    """
    try:
        if scalarize:
            return set(map(float, x.values))
        return x.values
    except Exception:
        raise OperationError(f'Object {str(x)} does not have a values property.')

def dim(x):
    "Returns the dimension of its argument, which is typically a kind, FRP, or statistic."
    try:
        return x.dim
    except Exception:
        raise OperationError(f'Object {str(x)} does not have a dim property.')

def codim(x):
    "Returns the co-dimension of its argument, which is typically a kind, FRP, or statistic."
    try:
        return x.codim
    except Exception:
        raise OperationError(f'Object {str(x)} does not have a codim property.')

def size(x):
    "Returns the size of its argument, which is typically a kind or FRP."
    try:
        return x.size
    except Exception:
        raise OperationError(f'Object {str(x)} does not have a size property.')

def clone(x):
    """Returns a clone of its argument, which is typically an FRP or conditional FRP.

    This operates on any object with a .clone() method, but is typically used
    to get a copy of an FRP or conditional FRP with the same kind but its
    own value.

    """
    try:
        return x.clone()
    except Exception as e:
        raise OperationError(f'Could not clone object {x}: {str(e)}')


#
# Tuples
#

def is_tuple(x: Any) -> TypeGuard[tuple]:
    "Is the given object a tuple?"
    return isinstance(x, tuple)

def scalarize(x):
    "If given a length 1 tuple, unwrap the value; otherwise returns argument as is."
    return (x[0] if isinstance(x, tuple) and len(x) == 1 else x)

def ensure_tuple(x: Any) -> tuple:
    "If given a non-tuple, wrap in a length 1 tuple; else returns argument as is."
    return (x if isinstance(x, tuple) else (x,))


#
# Sequences and Collections
#

def irange(
        start_or_stop: int,
        stop: int | None = None,
        *,
        step=1,
        exclude: Callable[[int], bool] | Iterable[int] | None = None,
        include: Callable[[int], bool] | Iterable[int] | None = None,
) -> Generator[int, None, None]:
    """Inclusive integer range.

    Parameters
    ----------
      start_or_stop - if the only argument, an integer giving the stop (inclusive)
          of the sequence; if stop is also supplied, this is the start.
      stop - if missing, start from 1 (unlike the builtin range that starts from 0);
          otherwise, the sequence goes up to and including this value.
      step - a non-zero integer giving the spacing between successive values of the
          sequence; it can e negativ if stop < start.
      exclude - either a set of integers or a predicate taking integers to boolean
          values; values in the set or for which the predicate returns true are skippe.
      include - either a set of integers or a predicate taking integers to boolean
          values; values in the set or for which the predicate returns true are included.
          If exclude is also supplied, this takes precedence.

    Returns an iterator over the resulting range.

    """
    if exclude is not None and not callable(exclude):
        exclude_values = set(exclude)
        exclude = lambda x: x in exclude_values
    if include is not None and not callable(include):
        include_values = set(include)
        include = lambda x: x in include_values

    if stop is None:
        stop = start_or_stop
        start = 1
    else:
        start = start_or_stop

    if (stop - start) * step < 0:
        raise ConstructionError(f'irange {start}:{stop} and step {step} have inconsistent direction.')

    sign = 1 if step >= 0 else -1

    def generate_from_irange() -> Generator[int, None, None]:
        value = start
        while (value - stop) * sign <= 0:
            if ((include is None and exclude is None) or
               (include is not None and include(value)) or
               (exclude is not None and not exclude(value))):
                yield value
            value += step

    return generate_from_irange()

def index_of(value, xs, not_found=-1, *, start=0, stop=sys.maxsize):
    """Returns index of `value` in `xs`, or `not_found` if none.

    If xs is a list or tuple, restrict attention to the slice
    from start to stop, exclusive, where start <= stop.

    """
    if stop <= start:
        return not_found

    if isinstance(xs, (list, tuple)):
        try:
            return xs.index(value, start, stop)
        except ValueError:
            return not_found
    else:
        for i, v in enumerate(xs):
            if i >= start and i < stop and v == value:
                return i
        return not_found

def frequencies(xs: Iterable[Hashable], counts_only=False) -> Union[dict[Hashable, int], tuple[int, ...]]:
    """Computes frequencies of the values in some iterable collection.

    If counts_only is False, returns a dictionary mapping the values to their counts.
    Otherwise, returns a tuple of counts in decreasing order.

    The items in the collection should be hashable.

    """
    freqs: dict[Hashable, int] = defaultdict(int)

    for x in xs:
        freqs[x] += 1

    if counts_only:
        return tuple(sorted(freqs.values(), reverse=True))
    return freqs


#
# Higher-Order Functions
#

def compose(*functions):
    """Returns a new function that composes the given functions successively.

    Note that compose(f,g) calls f *after* g. The values of g should be
    valid inputs to f, and similarly for any list of functions.
    This is not checked, however.

    """
    def compose2(f, g):
        return lambda x: f(g(x))

    n = len(functions)
    if n == 0:
        return identity

    if n == 1:
        return functions[0]

    if n == 2:
        return compose2(functions[0], functions[1])

    return reduce(compose2, functions)

    #  # For later Python versions
    # match functions:
    #     case ():
    #         return identity
    #     case (f,):
    #         return f
    #     case (f, g):
    #         return compose2(f, g)
    #     case _:
    #         return reduce(compose2, functions )

def lmap(func, *iterables):
    "Like the builtin `map` but automatically converts its results into a list."
    return list(map(func, *iterables))

def every(func, iterable):
    "Returns true if f(x) is truthy for every x in iterable."
    return all(map(func, iterable))

def some(func, iterable):
    "Returns true if f(x) is truthy for some x in iterable."
    return any(map(func, iterable))

def iterate(f: Callable[..., A], n: int, start: A, *extra_args, **extra_kwargs) -> A:
    """Returns nth item in the sequence: start, f(start), f(f(start)), f(f(f(start))), ...

    If n <= 0, start is returned as is. Extra positional and keyword arguments
    are passed to f in each call.

    Parameters
    ----------
    f :: a function from A -> A, however f can accept extra positional or keyword
        arguments; these are given by `extra_args` and `extra_kwargs`, respectively.

    n :: the number of times to iterate; if n <= 0, `start` is returned as is.

    start :: a value of type A, the initial value of the sequence

    extra_args :: zero or more additional arguments that are passed to `f`
        following the value of type A.

    extra_kwargs :: zero or more additional keyword arguments that are passed
        to `f` following the value of type A and `extra_args`.

    """
    result = start
    for _ in range(n):
        result = f(result, *extra_args, **extra_kwargs)
    return result


#
# Environment
#

def is_interactive() -> bool:
    "Checks if frp is running as an interactive app."
    return environment.is_interactive or hasattr(sys, 'ps1') or bool(sys.flags.interactive)

# ATTN: This needs work but is still useful
def show(x, *, print_it=True, indent=0, render=True):
    "Shows nested objects in the REPL in a more presentable fashion."
    if render and isinstance(x, Renderable):
        out = x.__frplib_repr__()
    elif isinstance(x, list):
        ind0 = (" " * indent)
        ind = ind0 + "  "
        sep = "\n" + ind
        init = "[\n" + ind
        final = "\n" + ind0 + "]"
        # Note: handle panels and block text
        out = init + sep.join([show(xi, print_it=False, indent=indent + 2, render=False)
                               for xi in x]) + final
    elif isinstance(x, dict):
        ind0 = (" " * indent)
        out = str({k: show(v, print_it=False) for k, v in x.items()})
    else:
        ind0 = (" " * indent)
        out = ind0 + str(x)
    if print_it:
        environment.console.print(out)
        return
    return out


#
# Info tags
#

setattr(clone, '__info__', 'utilities')
setattr(dim, '__info__', 'utilities')
setattr(codim, '__info__', 'utilities')
setattr(size, '__info__', 'utilities')
setattr(values, '__info__', 'utilities')
setattr(identity, '__info__', 'utilities')
setattr(const, '__info__', 'utilities')
setattr(compose, '__info__', 'utilities')
setattr(irange, '__info__', 'utilities::irange')
setattr(index_of, '__info__', 'utilities::index_of')
setattr(every, '__info__', 'utilities')
setattr(some, '__info__', 'utilities')
setattr(lmap, '__info__', 'utilities')
setattr(is_tuple, '__info__', 'utilities')
setattr(frequencies, '__info__', 'utilities')
setattr(show, '__info__', 'utilities::show')
setattr(iterate, '__info__', 'utilities::iterate')



# Hunter's Success Example in Chapter 0 Sec 8

from random              import random

from frplib.calculate    import substitute_with
from frplib.frps         import independent_mixture
from frplib.kinds        import (Kind, conditional_kind, fast_mixture_pow,
                                 constant, either, permutations_of, uniform, weighted_as,
                                 without_replacement)
from frplib.statistics   import (MonoidalStatistic, __, statistic,
                                 Abs, And, Diff, Fork, Min, Max, Or, Proj, Sqrt, Sum,)
from frplib.symbolic     import symbol, symbols
from frplib.utils        import clone, index_of, irange
from frplib.vec_tuples   import as_scalar, as_vec_tuple, vec_tuple

# Approach #1 in the Text is obtained with the profile expert_plus below

h = symbols('h_0 ... h_7')
hits = [weighted_as(0, 1, weights=[1 - w, w]) for w in h]

all_50_50 = substitute_with(dict(h_0=0.5, h_1=0.5, h_2=0.5, h_3=0.5,
                                 h_4=0.5, h_5=0.5, h_6=0.5, h_7=0.5))
one_expert = substitute_with(dict(h_0=0.1, h_1=0.1, h_2=0.1, h_3=0.1,
                                  h_4=0.1, h_5=0.1, h_6=0.1, h_7=0.9))
expert_plus = substitute_with(dict(h_0=0.1, h_1=0.1, h_2=0.1, h_3=0.1,
                                   h_4=0.1, h_5=0.1, h_6=0.7, h_7=0.9))
some_elders = substitute_with(dict(h_0=0.1, h_1=0.1, h_2=0.1, h_3=0.1,
                                   h_4=0.1, h_5=0.75, h_6=0.75, h_7=0.75))
random_skill = substitute_with({str(w): random() for w in h})

p = symbol('p')
all_equal = substitute_with(dict(h_0=p, h_1=p, h_2=p, h_3=p,
                                 h_4=p, h_5=p, h_6=p, h_7=p))

all_hits = independent_mixture(hits)
number_of_hits = Sum(all_hits)



from __future__ import annotations

import pytest

from frplib.exceptions import EvaluationError, ConstructionError, KindError
from frplib.kinds      import (Kind, kind, conditional_kind,
                               constant, either, uniform,
                               symmetric, linear, geometric,
                               weighted_by, weighted_as, arbitrary,
                               integers, evenly_spaced, bin,
                               subsets, permutations_of,
                               sequence_of_values,
                               fast_mixture_pow)
from frplib.numeric    import as_numeric, numeric_log2
from frplib.quantity   import as_quantity
from frplib.statistics import Proj, Sum, Min, Max
from frplib.symbolic   import symbol
from frplib.utils      import every, frequencies, irange, lmap
from frplib.vec_tuples import vec_tuple


def values_of(u):
    return u.keys()

def weights_of(u):
    return list(u.values())


def test_value_sequences():
    assert sequence_of_values(1, 2, 3) == [1, 2, 3]
    assert sequence_of_values(1) == [1]
    assert sequence_of_values(1, 2, ..., 6) == list(range(1, 7))
    assert sequence_of_values(10, 9, ..., 0) == list(range(10, -1, -1))
    assert sequence_of_values(0.05, 0.10, ..., 0.95, transform=as_quantity) == [as_quantity(0.05) * k for k in range(1, 20)]
    assert sequence_of_values(1, 2, ..., 1) == [1]
    assert sequence_of_values(1, 0, ..., 1) == [1]
    assert sequence_of_values(1, 1, ..., 1) == [1]
    assert sequence_of_values(1, 2, ..., 2) == [1, 2]
    assert sequence_of_values(1, 0, ..., 0) == [1, 0]

    with pytest.raises(KindError):
        sequence_of_values(1, 0, ..., 2)

    with pytest.raises(KindError):
        sequence_of_values(1, 2, ...)

    with pytest.raises(ConstructionError):
        sequence_of_values(1, 2, ..., 'a')

    with pytest.raises(KindError):
        sequence_of_values(1, 2, ..., 0)

    with pytest.raises(KindError):
        sequence_of_values(1, 2, ..., 1e9)

def test_kinds_factories():
    "Builtin kind factories"
    a = symbol('a')

    assert constant(1).values == {1}
    assert constant((2,)).values == {2}
    assert constant((2, 3)).values == {vec_tuple(2, 3)}

    assert either(0, 1).values == {0, 1}
    assert weights_of(either(0, 1, 2).weights) == pytest.approx([as_quantity('2/3'), as_quantity('1/3')])
    assert lmap(str, values_of(either(a, 2 * a, 2).weights)) == ['<a>', '<2 a>']

    u = uniform(1, 2, 3).weights
    assert values_of(u) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert weights_of(u) == pytest.approx([as_quantity('1/3'), as_quantity('1/3'), as_quantity('1/3')])

    w = weighted_as(1, 2, 3, weights=[1, 2, 4]).weights
    assert values_of(w) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert weights_of(w) == pytest.approx([as_quantity('1/7'), as_quantity('2/7'), as_quantity('4/7')])

    w = weighted_as(1, 2, 3, weights=[a, 2 * a, 4 * a]).weights
    assert values_of(w) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert weights_of(w) == pytest.approx([as_quantity('1/7'), as_quantity('2/7'), as_quantity('4/7')])

    w = weighted_as(1, 2, 3, weights=[1, 2 * a, 4 * a]).weights
    assert values_of(w) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert lmap(str, weights_of(w)) == ['1/(1 + 6 a)', '2 a/(1 + 6 a)', '4 a/(1 + 6 a)']

    w = weighted_as(a, 2 * a, 3 * a, weights=[1, 2, 4]).weights
    assert lmap(str, values_of(w)) == ['<a>', '<2 a>', '<3 a>']
    assert weights_of(w) == pytest.approx([as_quantity('1/7'), as_quantity('2/7'), as_quantity('4/7')])

def test_mixtures():
    k0 = either(10, 20)
    m0 = {10: either(4, 8, 99), 20: either(8, 4, 99)}
    m1 = conditional_kind(m0)
    me1 = {10: either(4, 8, 99), 30: either(8, 4, 99)}
    me2 = {10: either(4, 8, 99), (20, 30): either(8, 4, 99)}
    mec1 = conditional_kind(me1)
    mec2 = conditional_kind(me2)

    mix = (k0 >> m1).weights
    assert weights_of(mix) == pytest.approx([as_quantity('0.495'),
                                             as_quantity('0.005'),
                                             as_quantity('0.005'),
                                             as_quantity('0.495')])

    assert values_of(mix) == {vec_tuple(10, 4),
                              vec_tuple(10, 8),
                              vec_tuple(20, 4),
                              vec_tuple(20, 8),
                              }

    mix = (k0 >> m0).weights
    assert weights_of(mix) == pytest.approx([as_quantity('0.495'),
                                             as_quantity('0.005'),
                                             as_quantity('0.005'),
                                             as_quantity('0.495')])

    assert values_of(mix) == {vec_tuple(10, 4),
                              vec_tuple(10, 8),
                              vec_tuple(20, 4),
                              vec_tuple(20, 8),
                              }

    with pytest.raises(KindError):
        k0 >> me1

    with pytest.raises(KindError):
        k0 >> me2

    with pytest.raises(KindError):
        k0 >> mec1

    with pytest.raises(KindError):
        k0 >> mec2

    k1 = k0 >> m1 | (Proj[2] == 8)
    assert weights_of(k1.weights) == pytest.approx([as_quantity('0.01'), as_quantity('0.99')])
    assert values_of(k1.weights) == {vec_tuple(10, 8), vec_tuple(20, 8)}

    has_disease = either(0, 1, 999)     # No disease has higher weight
    test_by_status = conditional_kind({
        vec_tuple(0): either(0, 1, 99),     # No disease, negative has high weight
        vec_tuple(1): either(0, 1, '1/19')  # Yes disease, positive higher weight
    })

    dStatus_and_tResult = has_disease >> test_by_status
    Disease_Status = Proj[1]
    Test_Result = Proj[2]

    has_disease_updated = (dStatus_and_tResult | (Test_Result == 1))[Disease_Status]

    w = dStatus_and_tResult.weights
    assert values_of(w) == {vec_tuple(0, 0), vec_tuple(0, 1), vec_tuple(1, 0), vec_tuple(1, 1)}
    assert weights_of(w) == pytest.approx([as_quantity(v)
                                           for v in ['98901/100000', '999/100000', '1/20000', '19/20000']])

    w = has_disease_updated.weights
    assert values_of(w) == { vec_tuple(0), vec_tuple(1) }
    assert weights_of(w) == pytest.approx([as_quantity(v) for v in ['999/1094', '95/1094']])

def test_tagged_kinds():
    k = either(0, 1) * either(2, 3) * either(4, 5)

    k1 = Sum @ k | (Proj[2] == 2)

    list(k1.weights.values()) == [as_quantity('1/4'), as_quantity('1/2'), as_quantity('1/4')]
    list(k1.weights.keys()) == [vec_tuple(6), vec_tuple(7), vec_tuple(8)]

    k2 = Min @ k | (Proj[2] == 2)

    list(k2.weights.values()) == [as_quantity('1/4'), as_quantity('1/2'), as_quantity('1/4')]
    list(k2.weights.keys()) == [vec_tuple(6), vec_tuple(8), vec_tuple(9)]

def test_comparisons():
    assert 'same' in Kind.compare(uniform(1, 2), either(1, 2))
    assert 'differ' in Kind.compare(uniform(1, 2), either(1, 3))
    assert 'differ' in Kind.compare(uniform(1, 2), weighted_as(1, 2, weights=[0.999, 1.001]))

    assert Kind.equal(uniform(1, 2), either(1, 2))
    assert not Kind.equal(uniform(1, 2), either(1, 3))
    assert not Kind.equal(uniform(1, 2), weighted_as(1, 2, weights=[0.999, 1.001]))

    assert Kind.divergence(uniform(0, 2), uniform(0, 2)) == 0
    assert Kind.divergence(uniform(0, 2), weighted_as(1, 2, weights=[0.999, 1.001])) == as_quantity('Infinity')
    assert Kind.divergence(uniform(1, 2), weighted_as(1, 2, weights=['1/4', '3/4'])) == \
        pytest.approx(as_numeric('1/2') - numeric_log2('1.5') / 2)   # type: ignore

def test_fast_pow():
    assert Kind.equal(fast_mixture_pow(Sum, either(0, 1), 0), constant(0))
    assert Kind.equal(fast_mixture_pow(Min, either(0, 1), 0), constant('infinity'))
    assert Kind.equal(fast_mixture_pow(Max, either(0, 1), 0), constant('-infinity'))
    assert Kind.equal(fast_mixture_pow(Sum, either(0, 1), 1), either(0, 1))
    assert Kind.equal(fast_mixture_pow(Sum, either(0, 1), 2), Sum(either(0, 1) * either(0, 1)))
    assert Kind.equal(fast_mixture_pow(Sum, either(0, 1), 5), Sum(either(0, 1) ** 5))

def test_factory_details():
    # Test for roundoff that was happening in the differences with ...
    a_values = [as_numeric(0.05) * k for k in irange(1, 19)]
    k = uniform(0.05, 0.1, ..., 0.95)
    assert Kind.equal(k, uniform(a_values), tolerance='1.0e-16')

    k1 = weighted_as({0.05: 1, 0.45: 2, 0.70: 3})
    k2 = weighted_as(0.05, 0.45, 0.70, weights=[1, 2, 3])
    assert Kind.equal(k1, k2, tolerance='1.0e-16')

def test_indexing():
    k = either(0, 1) * either(2, 3) * either(4, 5) * either(6, 7)
    assert Kind.equal(k[:2], either(0, 1))
    assert Kind.equal(k[:-1], either(0, 1) * either(2, 3) * either(4, 5))
    assert Kind.equal(k[-3:-1], either(2, 3) * either(4, 5))
    with pytest.raises(KindError):
        k[5]
    with pytest.raises(KindError):
        k[-5]
    assert Kind.equal(k[:1], Kind.empty)

def test_sampling():
    c = symbol('c')

    assert len(constant(1).sample(10)) == 10
    assert every(lambda x: x == 1, constant(1).sample(10))
    assert every(lambda x: x == c, constant(c).sample(10))
    assert set(either(0, 1).sample(100)) == {(0,), (1,)}

    with pytest.raises(EvaluationError):
        either(0, 1, c).sample(10)

    a, b = frequencies(either(0, 1).sample(20000), counts_only=True)
    assert a + b == 20_000
    assert abs(a - 10000) <= 250



#
# Traffic Light from A Dialogue on Systems and State
#

from enum import IntEnum

from frplib.frps       import conditional_frp, frp
from frplib.exceptions import FrplibException
from frplib.kinds      import Kind, conditional_kind, constant, uniform, weighted_as
from frplib.statistics import Fork, Id
from frplib.utils      import clone

class TrafficLight(IntEnum):
    GREEN = 0
    YELLOW = 1
    RED = 2

    def __next__(self):
        if self == TrafficLight.GREEN:
            return TrafficLight.YELLOW

        if self == TrafficLight.YELLOW:
            return TrafficLight.RED

        return TrafficLight.GREEN

# First Effort

change_on = {
    TrafficLight.GREEN:  1 / 30,
    TrafficLight.YELLOW: 1 / 5,
    TrafficLight.RED:    1 / 30,
}

@conditional_frp
def tick_light(state):
    color, ticks = state
    change_probability = change_on[color]

    next_kind = weighted_as(
        (color, ticks + 1),
        (next(color), 0),
        weights=[1 - change_probability, change_probability]
    )
    return frp(next_kind)

# Provisional
@conditional_kind
def tick_light_kind(state):
    color, ticks = state
    change_probability = change_on[color]

    return weighted_as(
        (color, ticks + 1),
        (next(color), 0),
        weights=[1 - change_probability, change_probability]
    )

start_any_color = uniform(change_on.keys()) ^ Fork(Id, 0)
start_green = constant(TrafficLight.GREEN, 0)

def n_ticks(n, InitialState):
    assert n >= 0, "Number of ticks must be non-negative."

    if isinstance(InitialState, Kind):
        InitialState = frp(InitialState)

    State = InitialState
    for _ in range(n):
        State = clone(tick_light) // State

    return State

# Provisional
def n_ticks_kind(n, initial_state):
    assert n >= 0, "Number of ticks must be non-negative."

    state = initial_state
    for _ in range(n):
        state = tick_light_kind // state

    return state


# Graph Example

def neighbors(n: int) -> list[int]:
    # Ordinarily would use a dict but this matches the text
    # changed from match to if... so it would run in python 3.9
    # match n:
    #     case 1:
    #         return [2, 3, 4, 5, 8]
    #     case 2:
    #         return [1, 3, 4]
    #     case 3:
    #         return [1, 2]
    #     case 4:
    #         return [1, 2]
    #     case 5:
    #         return [1, 6]
    #     case 6:
    #         return [5, 8]
    #     case 7:
    #         return []
    #     case 8:
    #         return [6, 1]
    #     case _:
    #         raise FrplibException("Invalid node in graph example")

    if n == 1:
        return [2, 3, 4, 5, 8]
    if n == 2:
        return [1, 3, 4]
    if n == 3:
        return [1, 2]
    if n == 4:
        return [1, 2]
    if n == 5:
        return [1, 6]
    if n == 6:
        return [5, 8]
    if n == 7:
        return []
    if n == 8:
        return [6, 1]
    raise FrplibException("Invalid node in graph example")


@conditional_kind
def next_node(node):
    adjacent = neighbors(node)
    if len(adjacent) == 0:
        return constant(node)
    return uniform(adjacent)

def random_walk(start, n_steps: int = 1):
    assert n_steps >= 0, "Number of steps must be >= 0."

    current = start
    for _ in range(n_steps):
        current = next_node // current

    return current

@conditional_kind
def next_node_visit8(state):
    "Version of next_node that tracks whether we visited node 8."
    node, visit, time = state
    adjacent = neighbors(node)
    if len(adjacent) == 0:
        return constant(state)

    node_kind = uniform(adjacent)
    if visit == 1:
        return node_kind ^ Fork(Id, visit, time)
    if node == 8:
        return node_kind ^ Fork(Id, 1, time + 1)
    return uniform(adjacent) ^ Fork(Id, 0, time + 1)



# Random Image Example in Chapter 0 Section 2.3

import math

from collections.abc     import Iterable
from itertools           import chain, repeat
from random              import randrange
from typing              import cast, Literal, Union
from typing_extensions   import TypeAlias, Unpack

from frplib.exceptions   import IndexingError, OperationError
from frplib.frps         import FRP, frp
from frplib.kinds        import weighted_as
from frplib.quantity     import as_quantity
from frplib.statistics   import Statistic, statistic, Fork, Id
from frplib.utils        import irange
from frplib.vec_tuples   import VecTuple, as_vec_tuple, vec_tuple

ImageData: TypeAlias = tuple[Literal[0, 1], ...]
Image: TypeAlias = tuple[int, int, Unpack[ImageData]]
ImageId: TypeAlias = Union[str, int]
ModelId: TypeAlias = Union[str, int]

#
# Basic Image Components
#

def empty_image(width=32, height=32):
    "Returns an empty width x height image as a value."
    n = width * height
    return as_vec_tuple([width, height] + [0] * n)

pixel0 = (0,)  # These can be combined with as_image as in the text
pixel1 = (1,)


#
# Creating Images
#

def as_image(pixels: Iterable[Literal[0, 1]], width=32, height=32) -> Image:
    # ATTN: check for array structure, pad to proper length, etc
    return cast(Image, vec_tuple(width, height, *pixels))


#
# Helpers
#

def conform_image(image: Image, width=32, height=32) -> Image:
    "ATTN"
    wd, ht = image[:2]

    # Optimize for the most common case at the cost of an extra comparison
    if wd == width and ht == height:
        return image

    data: ImageData = image[2:]   # type: ignore

    if wd == width:
        if ht > height:
            return as_image(data, width, height)
        return as_image(chain(data, repeat(0, width * (height - ht))), width, height)

    n = width * height
    conf: list[Literal[0, 1]] = [0] * n

    if wd < width:
        for i in range(min(ht, height)):
            conf[(i * width):(i * width + wd)] = data[(i * wd):((i + 1) * wd)]
    else:
        for i in range(min(ht, height)):
            conf[(i * width):((i + 1) * width)] = data[(i * wd):(i * wd + width)]

    return as_image(conf, width, height)

def ensure_same_dims(image1: Image, image2: Image) -> tuple[int, int]:
    "Ensures that two images have the same dimensions."
    wd1, ht1 = image1[:2]
    wd2, ht2 = image2[:2]

    if wd1 != wd2:
        raise OperationError(f'Attempt to add images of different widths {wd1} != {wd2}')

    if ht1 != ht2:
        raise OperationError(f'Attempt to add images of different heights {ht1} != {ht2}')

    return (wd1, ht1)

def image_data(image: Image) -> tuple[int, int, ImageData]:
    "Decomposes an encoded image into dimensions and binary image data."
    wd, ht = image[:2]
    data = cast(ImageData, image[2:])
    return (wd, ht, data)

def add_base(base: Image):
    ""
    wd, ht = base[:2]
    n = wd * ht

    @statistic
    def do_add(img):
        # ensure_same_dims(img, base)  # conform image to base here but for now...
        new_img = cast(VecTuple, conform_image(img, wd, ht))
        return as_image((new_img[2 + i] ^ base[2 + i] for i in range(n)), wd, ht)

    return do_add

#
# Manipulating Images
#

def add_image(image1: Image, image2: Image) -> Image:
    #
    wd, ht = ensure_same_dims(image1, image2)
    n = wd * ht
    return as_image((image1[2 + i] ^ image2[2 + i] for i in range(n)), wd, ht)   # type: ignore

def clockwise(image: Image) -> Image:
    "A statistic that rotates an image 90 degrees clockwise."
    wd, ht, data = image_data(image)

    rotated: list[Literal[0, 1]] = [0] * (wd * ht)
    for j in range(ht):
        for i in range(wd):
            rotated[(ht - j - 1) + i * ht] = data[i + j * wd]

    return as_image(rotated, ht, wd)

def counter_clockwise(image: Image) -> Image:
    "A statistic that rotates an image 90 degrees counter-clockwise."
    wd, ht, data = image_data(image)

    rotated: list[Literal[0, 1]] = [0] * (wd * ht)
    for j in range(ht):
        for i in range(wd):
            rotated[j + (wd - i - 1) * ht] = data[i + j * wd]

    return as_image(rotated, wd, ht)

@statistic
def reflect_image_horizontally(image: Image) -> Image:
    "A statistic that reflects an image across its vertical midline."
    wd, ht, data = image_data(image)

    reflected: list[Literal[0, 1]] = [0] * (wd * ht)
    for j in range(ht):
        for i in range(wd):
            reflected[(wd - i - 1) + j * wd] = data[i + j * wd]

    return as_image(reflected, wd, ht)

@statistic
def reflect_image_vertically(image: Image) -> Image:
    "A statistic that reflects an image across its horizontal midline."
    wd, ht, data = image_data(image)

    reflected: list[Literal[0, 1]] = [0] * (wd * ht)
    for j in range(ht):
        for i in range(wd):
            reflected[i + (ht - j - 1) * wd] = data[i + j * wd]

    return as_image(reflected, wd, ht)

def largest_cluster_size(image: Image):
    ...

#
# Main Image FRP Factory
#

def random_image(p='1/2', base: Union[Image, None] = None, width=None, height=None) -> FRP:
    """Returns an FRP representing a width x height random binary image.

    The image is represented as a tuple stored row-wise from the top
    left to the bottom right of the image.

    ATTN

    """
    if width is None:
        width = 32 if base is None else base[0]

    if height is None:
        height = 32 if base is None else base[1]

    p = as_quantity(p)
    pixel = weighted_as(0, 1, weights=[1 - p, p])
    n = width * height

    noise: FRP = (frp(pixel) ** n) ^ Fork(width, height, Id)
    if base is None:
        return noise

    shift = add_base(conform_image(base, width, height))
    return shift(noise)


#
# Simple Image Statistics
#

@statistic
def black_pixels(image: Image):
    "Statistic that counts the number of black pixels in an image."
    _, _, data = image_data(image)
    return sum(data)


#
# Erosion and Dilation
#

def erode(element: Union[int, Iterable[tuple[int, int]]] = 1) -> Statistic:
    ""
    if isinstance(element, int):
        s = abs(element)
        delta_xl = s
        delta_xr = s
        delta_yl = s
        delta_yu = s
        element = set([(i, j) for i in irange(-s, s) for j in irange(-s, s)])
    else:
        element = set(element)
        delta_xl = max((abs(xy[0]) if xy[0] < 0 else 0) for xy in element)
        delta_xr = max((xy[0] if xy[0] > 0 else 0) for xy in element)
        delta_yl = max((abs(xy[1]) if xy[1] < 0 else 0) for xy in element)
        delta_yu = max((xy[1] if xy[1] > 0 else 0) for xy in element)

    @statistic
    def erosion(image: Image):
        wd, ht, data = image_data(image)
        width = wd - delta_xl - delta_xr
        height = ht - delta_yl - delta_yu
        eroded: list[Literal[0, 1]] = [0] * (width * height)

        for y in range(height):
            for x in range(width):
                if all(data[x + i + delta_xl + wd * (y + j + delta_yl)] == 1 for i, j in element):
                    eroded[x + width * y] = 1

        return as_image(eroded, width, height)

    return cast(Statistic, erosion)

def dilate(element: Union[int, Iterable[tuple[int, int]]] = 1) -> Statistic:
    ""
    if isinstance(element, int):
        s = abs(element)
        delta_xl = s
        delta_xr = s
        delta_yl = s
        delta_yu = s
        element = set([(i, j) for i in irange(-s, s) for j in irange(-s, s)])
    else:
        element = set(element)
        delta_xl = max((abs(xy[0]) if xy[0] < 0 else 0) for xy in element)
        delta_xr = max((xy[0] if xy[0] > 0 else 0) for xy in element)
        delta_yl = max((abs(xy[1]) if xy[1] < 0 else 0) for xy in element)
        delta_yu = max((xy[1] if xy[1] > 0 else 0) for xy in element)

    @statistic
    def dilation(image):
        wd, ht, data = image_data(image)
        width = wd + delta_xl + delta_xr
        height = ht + delta_yl + delta_yu
        dilated: list[Literal[0, 1]] = [0] * (width * height)

        for y in range(ht):
            for x in range(wd):
                if data[x + y * wd] == 1:
                    for i, j in element:
                        dilated[x + i + delta_xl + width * (y + j + delta_yl)] = 1

        return as_image(dilated, width, height)

    return cast(Statistic, dilation)


#
# Image Sets
#

class ImageSet:
    """
    ATTN

    """
    def __init__(self):
        self._models: dict[ModelId, list[Image]] = {}
        self._images: dict[ImageId, Image] = {}

    def register_image(self, image_id: ImageId, image: Image) -> None:
        self._images[image_id] = image

    def register_model(self, model_id: ModelId, images: Iterable[Image]) -> None:
        self._models[model_id] = list(images)

    def images(self) -> list[ImageId]:
        return list(self._images.keys())

    def image(self, image_id: ImageId) -> Image:
        if image_id in self._images:
            return self._images[image_id]
        raise IndexingError(f'Unknown image {image_id} in image registry')

    def models(self) -> list[ModelId]:
        return list(self._models.keys())

    def model(self, model_id: ModelId) -> list[Image]:
        if model_id in self._models:
            return self._models[model_id]
        raise IndexingError(f'Unknown model {model_id} in model registry')

    def observe(self, model_id: Union[ModelId, list[Image]], p='1/8'):
        "ATTN"
        if isinstance(model_id, (str, int)):
            model = self._models[model_id]
        else:
            model = model_id

        truth = model[randrange(len(model))]  # select a base uniformly
        data = random_image(base=truth, p=as_quantity(p))

        return (data, truth)

# Predefined Images and Models

ImageModels = ImageSet()

ImageModels.register_image(
    'simple',
    as_image([1 if ((i == 0 and j == 0) or
                    (i == 31 and j == 0) or
                    (i == 0 and j == 31) or
                    (i == 31 and j == 31) or
                    (i == 15 and j == 0) or
                    (i == 17 and j == 0) or
                    (i == 15 and j == 31) or
                    (i == 17 and j == 31) or
                    (i == 0 and j == 15) or
                    (i == 0 and j == 17) or
                    (i == 31 and j == 15) or
                    (i == 31 and j == 17) or
                    (i == 4 and j == 4) or
                    (i == 28 and j == 4) or
                    (i == 4 and j == 28) or
                    (i == 28 and j == 28) or
                    (8 <= i <= 9 and 8 <= j <= 9) or
                    (23 <= i <= 24 and 8 <= j <= 9) or
                    (8 <= i <= 9 and 23 <= j <= 24) or
                    (23 <= i <= 24 and 23 <= j <= 24) or
                    (14 <= i <= 18 and 14 <= j <= 18))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)


ImageModels.register_image(
    'blocks',
    as_image([1 if ((4 <= i <= 8 and 4 <= j <= 8) or
                    (6 <= i <= 12 and 10 <= j <= 13) or
                    (18 <= i <= 24 and 18 <= j <= 24) or
                    (28 <= i <= 30 and 9 <= j <= 21) or
                    (2 <= i <= 7 and 25 <= j <= 30) or
                    (28 <= i <= 29 and 28 <= j <= 29) or
                    (25 <= i <= 26 and 28 <= j <= 29) or
                    (i == 16 and j == 9) or
                    (11 <= i <= 12 and 18 <= j <= 27) or
                    (12 <= i <= 30 and 1 <= j <= 2))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'squares',
    as_image([1 if ((2 <= i <= 3 and 2 <= j <= 3) or
                    (5 <= i <= 6 and 2 <= j <= 3) or
                    (2 <= i <= 3 and 5 <= j <= 6) or
                    (5 <= i <= 6 and 5 <= j <= 6) or
                    (9 <= i <= 10 and 2 <= j <= 3) or
                    (14 <= i <= 15 and 2 <= j <= 3) or
                    (17 <= i <= 19 and 2 <= j <= 4) or
                    (21 <= i <= 23 and 2 <= j <= 4) or
                    (26 <= i <= 28 and 2 <= j <= 4) or
                    (2 <= i <= 5 and 8 <= j <= 11) or
                    (8 <= i <= 11 and 8 <= j <= 11) or
                    (13 <= i <= 16 and 8 <= j <= 11) or
                    (19 <= i <= 22 and 8 <= j <= 11) or
                    (27 <= i <= 30 and 8 <= j <= 11) or
                    (8 <= i <= 11 and 13 <= j <= 16) or
                    (13 <= i <= 16 and 14 <= j <= 17) or
                    (20 <= i <= 24 and 20 <= j <= 24) or
                    (27 <= i <= 31 and 20 <= j <= 24) or
                    (14 <= i <= 19 and 27 <= j <= 32) or
                    (i == 22 and j == 27) or
                    (i == 24 and j == 27) or
                    (i == 27 and j == 27) or
                    (i == 31 and j == 27) or
                    (i == 22 and j == 29) or
                    (i == 24 and j == 29) or
                    (i == 27 and j == 29) or
                    (i == 31 and j == 29) or
                    (i == 22 and j == 32) or
                    (i == 24 and j == 32) or
                    (i == 27 and j == 32) or
                    (i == 31 and j == 32) or
                    (2 <= i <= 9 and 21 <= j <= 28))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'minus',
    as_image([1 if (8 <= i <= 23 and 15 <= j <= 17) else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'P',
    as_image([1 if ((4 <= i <= 8 and 4 <= j <= 28) or
                    (9 <= i <= 17 and 4 <= j <= 8) or
                    (18 <= i <= 22 and 4 <= j <= 16) or
                    (9 <= i <= 17 and 12 <= j <= 16) or
                    (25 <= i <= 30 and 25 <= j <= 30))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'E',
    as_image([1 if ((4 <= i <= 7 and 4 <= j <= 28) or
                    (8 <= i <= 17 and 4 <= j <= 7) or
                    (8 <= i <= 17 and 15 <= j <= 17) or
                    (8 <= i <= 17 and 26 <= j <= 28))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'F',
    as_image([1 if ((4 <= i <= 7 and 4 <= j <= 28) or
                    (8 <= i <= 17 and 4 <= j <= 7) or
                    (8 <= i <= 17 and 15 <= j <= 17))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'F#',
    as_image([1 if ((16 <= i <= 28 and 16 <= j <= 112) or
                    (29 <= i <= 68 and 16 <= j <= 28) or
                    (29 <= i <= 68 and 60 <= j <= 72))
              else 0 for j in range(128) for i in range(128)],
             128, 128)
)

ImageModels.register_image(
    'H',
    as_image([1 if ((4 <= i <= 7 and 4 <= j <= 28) or
                    (8 <= i <= 17 and 15 <= j <= 17) or
                    (17 <= i <= 20 and 4 <= j <= 28))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_image(
    'L',
    as_image([1 if ((4 <= i <= 7 and 4 <= j <= 28) or
                    (8 <= i <= 17 and 25 <= j <= 28) or
                    (25 <= i <= 30 and 5 <= j <= 10))
              else 0 for j in range(32) for i in range(32)],
             32, 32)
)

ImageModels.register_model('efh', [ImageModels.image('E'), ImageModels.image('F'), ImageModels.image('H')])

#
# Image Utilities
#

def image_distance(image1: Image, image2: Image) -> int:
    "Returns the Hamming distance between two binary images of the same dimensions."
    wd, ht = ensure_same_dims(image1, image2)
    n = wd * ht
    return sum(image1[2 + i] ^ image2[2 + i] for i in range(n))  # type: ignore

def closest_image_to(image: Image, candidates: Iterable[Image]) -> Union[Image, None]:
    best, min_dist = None, None

    for cand_img in candidates:
        dist = image_distance(image, cand_img)
        if (min_dist is None) or (dist < min_dist):   # type: ignore
            min_dist = dist
            best = cand_img

    return best


#
# Reconstruction Methods
#

def reconstruct_image(model_id, denoiser=dilate()(erode())):
    """A statistics factory for reconstructing an unknown image from noisy data.

    Parameters
    ----------
    model_id: int | str - an identifier for the model holding candidate base images
    denoiser: Statistic - a denoising statistic mapping image to image

    """
    model_images = ImageModels.model(model_id)

    @statistic
    def reconstruct(observed_image):
        denoised_image = denoiser(observed_image)
        return closest_image_to(denoised_image, model_images)

    return reconstruct

def max_likelihood_image(
        model_id: ModelId,
        return_p=False,
        delta_p=0.001,
        end_p=0.5,
        start_p=0.001
) -> Statistic:
    """ATTN

    """

    def log_like(noise, log_p, log_1_p):
        wd, ht, data = image_data(noise)
        ones = sum(data)
        return log_p * ones + log_1_p * (wd * ht - ones)

    models = cast(list[VecTuple], ImageModels.model(model_id))

    @statistic
    def ml_recon(noise_like):
        max_ll = -math.inf
        best_p = 0
        best_m = None

        p = start_p
        while p < end_p:
            log_p = math.log(p)
            log_1_p = math.log(1 - p)
            for m in models:
                ll = log_like(add_image(cast(Image, m), noise_like), log_p, log_1_p)
                if ll > max_ll:
                    max_ll = ll
                    best_p = p
                    best_m = m
            p += delta_p

        if best_m is None:
            raise OperationError('ATTN')

        if return_p:
            img = list(best_m)
            img.append(best_p)
            return VecTuple(img)
        return best_m

    return cast(Statistic, ml_recon)

#
# Reconstruction Simulator
#

def simulate_denoise(
        model_id: ModelId,
        denoiser,
        p='1/8',
        observations=10_000
):
    """Evaluates denoising statistic on repeated observations from a model.

    Parameters:
     + model_id: ModelId - identifier of pre-defined model in ImageModels
     + denoiser: Statistic - a denoising statistic mapping image to image
     + p: ScalarQ - noise prevalance (0 <= p <= 1), numeric or string
     + observations: int - number of observed images to generate

    Returns a pair of numbers giving (i) the proportion of incorrect
    reconstructions over all observations, and (ii) the average
    distance between truth and reconstruction over all observations.

    """
    propi = 0
    score = 0
    for _ in range(observations):
        data, truth = ImageModels.observe(model_id, p=p)
        reconstructed = denoiser(data)
        distance = image_distance(reconstructed.value, truth)
        score += distance
        propi += (distance > 0)  # 0 if correct, 1 if not
    return (propi / observations, score / observations)


#
# Utility for making image files (not for general use)
#

def mvg_image(image: Image) -> str:
    "Returns mvg format description of an image."
    wd, ht = image[:2]
    data = cast(ImageData, image[2:])

    points = [f'push graphic-context\n  viewbox 0 0 {wd} {ht}']
    for i, px in enumerate(data):
        if px == 1:
            points.append(f'  point {i % wd},{i // wd}')
    points.append('pop graphic-context')
    return '\n'.join(points) + '\n'

def mvg_of(img: Union[Image, FRP]) -> str:
    "Returns mvg format description of image or image FRP."
    if isinstance(img, FRP):
        img = img.value      # type: ignore
    return mvg_image(img)    # type: ignore

def mvg_out(filename: str, img: Union[Image, FRP]) -> None:
    "Output image in mvg format to a specified file."
    with open(filename, 'w') as f:
        f.write(mvg_of(img))



from __future__ import annotations

class FrplibException(Exception):
    "Base exception for specialized playground user errors."
    pass

class FrplibInternalException(FrplibException):
    "Base exception for internal condtions in the library."
    pass

class MismatchedDomain(FrplibInternalException):
    "A value outside the expected domain was passed to a mapping."
    pass

class DomainDimensionError(FrplibInternalException):
    "A value outside the expected domain was passed to a mapping."
    pass

class ConstructionError(FrplibInternalException):
    "A problem was encountered creating an object."
    pass

class EvaluationError(FrplibInternalException):
    "A problem was encountered while evaluating input or derived values."
    pass

class KindError(FrplibException):
    "An error encountered in creating or using a kind."
    pass

class FrpError(FrplibException):
    "An error encountered in creating or using an FRP."
    pass

class MarketError(FrplibException):
    "An error encountered during market interactive session."
    pass

class PlaygroundError(FrplibException):
    "An error encountered during playground interactive session."
    pass

class OperationError(PlaygroundError):
    "An error encountered during a mathematical operation"
    pass

class ConditionMustBeCallable(ConstructionError):
    "A non-callable object cannot be converted to a condition."
    pass

class StatisticError(ConstructionError):
    "A problem with creating a Statistic.."
    pass

class IndexingError(PlaygroundError):
    "An error encountered during a key lookup or indexing operation"
    pass

class ComplexExpectationWarning(FrplibException):
    "Warning: Computing an expectation may be computationally inadvisable."
    pass

class NumericConversionError(FrplibException):
    "A quantity could not be converted to numeric form."
    pass



from __future__ import annotations

import re

from abc               import ABC
from collections       import defaultdict
from collections.abc   import Generator
from decimal           import Decimal
from functools         import reduce
from operator          import add
from typing            import cast, Literal, Union
from typing_extensions import TypeGuard

from frplib.exceptions import ConstructionError
from frplib.numeric    import (Numeric, ScalarQ,
                               as_real, as_nice_numeric, as_numeric,
                               is_scalar_q, show_numeric)
# from frplib.protocols  import NamedCallable
# from frplib.utils      import every

#
# Helpers
#

def merge_with(a, b, merge_fn=lambda x, y: y):
    merged = {k: a.get(k, b.get(k)) for k in a.keys() ^ b.keys()}
    merged.update({k: merge_fn(a[k], b[k]) for k in a.keys() & b.keys()})
    return merged

def is_zero(x: Numeric):  # Move to Numeric, should it take ScalarQ?
    return x == 0  # Temp

def show_coef(x: ScalarQ) -> str:
    return show_numeric(as_numeric(x), max_denom=1)

def purify(x: Symbolic) -> Symbolic | Numeric:
    pv = x.pure_value()
    if pv is not None:
        return pv
    return x

def common_term_rf(acc: dict[str, int] | None, multi: SymbolicMulti) -> dict[str, int]:
    """Reducing function that extracts common multi terms from a multi sum. Coefs excluded.

    Apply this to a SymbolicMultiSum x with reduce(common_term_rf, x). This returns
    a term object for a SymbolicMulti with the variables and their powers represented.
    An empty object corresponds to a constant, i.e., no nontrivial common term.

    """
    if acc is None:
        acc = multi.term.copy()
        return acc
    acc_prime = defaultdict(int)
    for var in acc:
        if var in multi.term:
            acc_prime[var] = min(acc[var], multi.term[var])
    return acc_prime

def common_term(sym_sum: SymbolicMultiSum) -> tuple[SymbolicMulti, SymbolicMultiSum]:
    "Extracts a common term (if any) from sum and return (term, adjusted sum)."
    term = reduce(common_term_rf, sym_sum.terms, None)
    assert term is not None

    if len(term) == 0:
        return (symbolic_one, sym_sum)

    common = SymbolicMulti.from_terms(term, sym_sum.coef)
    adj_terms = []
    for oterm in sym_sum.terms:
        adj, _, _ = multis_divide(oterm, common)
        adj_terms.append(adj / sym_sum.coef)
    adj_sum = SymbolicMultiSum(adj_terms)

    return (common, adj_sum)

def multis_divide(
        multi_num: SymbolicMulti,
        multi_den: SymbolicMulti
) -> tuple[SymbolicMulti, SymbolicMulti, bool]:
    """
    Returns None if the terms do not divide, or (1, denominator) or (numerator, 1)
    if they do, where the non-1 term is the non-canceled remaining term.

    """
    dpv = multi_den.pure_value()
    if dpv is not None:
        return (multi_num / dpv, symbolic_one, False)

    npv = multi_num.pure_value()
    if npv is not None:
        return (symbolic_one, multi_den / npv, False)

    if (multi_den.term.keys() <= multi_num.term.keys() or
       multi_num.term.keys() < multi_den.term.keys()):
        num = multi_num.term.copy()
        den = multi_den.term.copy()

        for k in multi_num.term.keys() & multi_den.term.keys():
            if den[k] <= num[k]:
                num[k] -= den[k]
                del den[k]
            else:
                den[k] -= num[k]
                del num[k]

        return (SymbolicMulti.from_terms(num, multi_num.coef),
                SymbolicMulti.from_terms(den, multi_den.coef),
                True)

    return (multi_num, multi_den, False)

def factor_quadratic(multisum: SymbolicMultiSum):
    pass


#
# Unique Symbols
#

def symbol_name_generator(base='#x_') -> Generator[str, None, None]:
    script = 0
    while True:
        yield f'{base}{{{script}}}'
        script += 1

gensym = symbol_name_generator()


#
# Generic Symbolic Quantities
#

class Symbolic(ABC):
    def is_pure(self):
        ...

    def pure_value(self):
        ...

    @property
    def key_of(self):
        ...

    @property
    def sort_key(self):
        ...

    def substitute(self, mapping: dict[str, ScalarQ], purify=True):
        ...

    def __lt__(self, other):
        "Sort symbols before numbers and by sort_key against Symbolics."
        if isinstance(other, Symbolic):
            return self.sort_key < other.sort_key
        if isinstance(other, str):
            return str(self) < other
        if is_scalar_q(other):
            pv = self.pure_value()
            if pv is not None:
                return pv < as_real(other)
        return True

    def __gt__(self, other):
        "Sort symbols before numbers and by sort_key against Symbolics."
        if isinstance(other, Symbolic):
            return self.sort_key > other.sort_key
        if isinstance(other, str):
            return str(self) > other
        if is_scalar_q(other):
            pv = self.pure_value()
            if pv is not None:
                return pv > as_real(other)
        return False

    def __le__(self, other):
        "Sort symbols before numbers and by sort_key against Symbolics."
        if isinstance(other, Symbolic):
            return self.sort_key <= other.sort_key
        if isinstance(other, str):
            return str(self) <= other
        if is_scalar_q(other):
            pv = self.pure_value()
            if pv is not None:
                return pv <= as_real(other)
        return True

    def __ge__(self, other):
        "Sort symbols before numbers and by sort_key against Symbolics."
        if isinstance(other, Symbolic):
            return self.sort_key >= other.sort_key
        if isinstance(other, str):
            return str(self) >= other
        if is_scalar_q(other):
            pv = self.pure_value()
            if pv is not None:
                return pv >= as_real(other)
        return False


#
# Multinomial Terms
#

class SymbolicMulti(Symbolic):
    "A symbolic multinomial term c a_1^k_1 a_2^k_2 ... a_n^k_n."
    def __init__(self, vars: list[str], powers: list[int], coef: ScalarQ = 1):
        # if powers too short, those count as 0, so ok; extra powers ignored
        # coef, [], []  acts as a scalar and a *multiplicative* identity
        self.coef = as_numeric(coef)

        order = 0
        sig = '1'  # show_coef(self.coef)
        multi: dict[str, int] = defaultdict(int)
        if not is_zero(self.coef):
            for var, pow in zip(vars, powers):
                if not var:
                    raise ConstructionError('A symbolic variable name must be a non-empty string.')
                multi[var] += pow
                order += pow
            sigs = []
            for var, pow in sorted(multi.items()):
                if pow == 0:
                    del multi[var]
                    continue
                sigs.append(f'{var}^{pow}')
            if sigs:
                sig = " ".join(sigs)

        self.term = multi
        self.order = order
        self.key = sig
        self.as_str: Union[str, None] = None   # Computed lazily

    @classmethod
    def from_terms(cls, multi: dict[str, int], coef: Numeric = 1):
        if is_zero(coef):
            return SymbolicMulti([], [], 0)
        vars = []
        pows = []
        for var, pow in multi.items():
            if pow != 0:
                vars.append(var)
                pows.append(pow)
        return cls(vars, pows, coef)

    @classmethod
    def pure(cls, x: ScalarQ = 0):
        return cls([], [], as_numeric(x))

    def is_pure(self) -> bool:
        return len(self.term) == 0

    def pure_value(self):
        if self.is_pure():
            return self.coef
        return None

    def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> SymbolicMulti | Numeric:
        coef = self.coef
        multi = self.term.copy()
        for var, pow in self.term.items():
            if var in mapping:
                coef *= as_numeric(mapping[var]) ** pow
                del multi[var]
        coef = as_numeric(coef)  # ATTN: needed?
        if len(multi) == 0:
            return coef
        return SymbolicMulti.from_terms(multi, coef)

    def reset_coef(self, new_coef):
        return SymbolicMulti.from_terms(self.term, new_coef)

    @property
    def signature(self) -> str:
        return self.key

    @property
    def key_of(self):
        return self.key

    @property
    def sort_key(self):
        return (self.coef, self.key)

    def __bool__(self) -> bool:
        return False if self.is_pure() and is_zero(self.coef) else True

    def __str__(self) -> str:
        if self.as_str is None:
            if self.is_pure():
                self.as_str = show_coef(self.coef)
            else:
                coef = '' if self.coef == 1 else show_coef(self.coef) + ' '
                term = re.sub(r'\^1( |$)', lambda m: ' ' if m.group(1) else '', self.signature)
                self.as_str = coef + term
        return self.as_str

    def __frplib_repr__(self) -> str:
        return str(self)

    def clone(self) -> SymbolicMulti:
        return self.from_terms(self.term, self.coef)

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return 0
            return self.from_terms(self.term, self.coef * as_real(other))
        if isinstance(other, SymbolicMulti):
            term = other.term if self.is_pure() else merge_with(self.term, other.term, add)
            coef = self.coef * other.coef
            return SymbolicMulti.from_terms(term, coef)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return 0
            return self.from_terms(self.term, as_real(other) * self.coef)
        # Cannot be SymbolicMulti in rul
        return NotImplemented

    def __add__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, self.coef + as_real(other))
            return SymbolicMultiSum([self, SymbolicMulti.pure(as_numeric(other))])
        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([self, other])
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, as_real(other) + self.coef)
            return SymbolicMultiSum([SymbolicMulti.pure(as_numeric(other)), self])
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, self.coef - as_real(other))
            return SymbolicMultiSum([self, SymbolicMulti.pure(as_numeric(-other))])
        if isinstance(other, SymbolicMulti):
            if self.key_of == other.key_of:
                return as_numeric(self.coef - other.coef)
            return SymbolicMultiSum([self, -1 * other])
        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, as_real(other) - self.coef)
            return SymbolicMultiSum([SymbolicMulti.pure(as_numeric(other)), -1 * self])
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return self.from_terms(self.term, self.coef / as_real(other))

        if isinstance(other, (SymbolicMulti, SymbolicMultiSum)):
            return symbolic(self, other)

        if isinstance(other, SymbolicMultiRatio):
            return symbolic(self * other.denominator, other.numerator)

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return symbolic(SymbolicMulti.pure(other), self)

        if isinstance(other, (SymbolicMulti, SymbolicMultiSum)):
            return symbolic(other, self)

        if isinstance(other, SymbolicMultiRatio):
            return symbolic(other.numerator, self * other.denominator)

        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, int):
            if n == 0:
                return 1
            term = self.term.copy()
            for var, pow in term.items():
                term[var] = pow * n if var else pow   # Constant always has power 1
            return self.from_terms(term, self.coef ** n)
        return NotImplemented

symbolic_zero = SymbolicMulti.pure(0)
symbolic_one = SymbolicMulti.pure(1)


#
# Symbolic Function Calls
#

# class SymbolicFn(SymbolicMulti):
#     def __init__(self, fn: NamedCallable, args: list[Symbolic]) -> None:
#         var = next(gensym)
#         self.fns = {var: {'fn': fn, 'args': args[:]}}
#         super().__init__([var], [1])
#
#     def is_pure(self) -> bool:
#         return (super().is_pure() or
#                 (self.term.keys() == self.fns.keys() and
#                  all(every(lambda s: s.is_pure(), f['args']) for f in self.fns.values())))
#
#     def pure_value(self):
#         if self.is_pure():
#             value = self.coef
#             for fn_key in self.fns:
#                 fstar = self.fns[fn_key]
#                 fn = cast(Callable[..., Numeric], fstar['fn'])
#                 args: list = fstar['args']
#                 assert callable(fn)
#                 value *= as_real(fn(*args))
#             return value
#         return None
#
#     def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> SymbolicFn | Numeric:
#         multi = self.substitute(mapping, purify)
#         pass
#
#     def __mul__(self, other):
#         # Stay as SymbolicFn against SymbolicMulti and pure
#         pass
#
#     def __rmul__(self, other):
#         # Stay as SymbolicFn against SymbolicMulti and pure
#         pass


#
# Sums of Multinomial Terms
#

class SymbolicMultiSum(Symbolic):
    """A sum of symbolic multinomial terms sum_i c_i a_i1^k_1 a_i2^k_2 ... a_in^k_n.

    """
    def __init__(self, multis: list[SymbolicMulti]) -> None:
        terms = self.combine_terms(multis)
        if not terms:
            self.terms: list[SymbolicMulti] = []
            self.coef: Numeric = 0
            self.order = 0
            self.key = '0'
            self.as_str: str | None = '0'
        elif len(terms) == 1 and terms[0].is_pure():
            self.terms = [terms[0]]
            self.coef = terms[0].pure_value()
            self.order = 0
            self.key = '1'
            self.as_str = show_coef(self.coef)
        else:
            terms = sorted(terms, key=lambda t: t.signature)
            self.order = 0
            for term in terms:
                if term.order > self.order:
                    self.order = term.order
                    self.coef = term.coef
            self.key = " + ".join([str(term / self.coef) for term in terms])
            self.terms = terms
            self.as_str = None

    def __str__(self) -> str:
        if self.as_str is None:
            self.as_str = " + ".join([str(term) for term in self.terms])
        return self.as_str

    def __frplib_repr__(self) -> str:
        return str(self)

    def is_pure(self) -> bool:
        return len(self.terms) == 1 and self.terms[0].is_pure()

    def is_single(self) -> bool:
        return len(self.terms) == 1

    def pure_value(self):
        if self.is_pure():
            return self.coef
        return None

    def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> SymbolicMultiSum | Numeric:
        if self.is_pure() and purify:
            return as_nice_numeric(self.coef)
        total: Union[SymbolicMultiSum, Numeric] = sum(term.substitute(mapping, purify=purify)  # type: ignore
                                                      for term in self.terms)
        if isinstance(total, Symbolic):
            return total
        return as_nice_numeric(total)

    @property
    def key_of(self):
        return self.key

    @property
    def sort_key(self):
        return (self.coef, self.key)

    @classmethod
    def singleton(cls, sym: SymbolicMulti):
        return SymbolicMultiSum([sym])

    @staticmethod
    def combine_terms(terms: list[SymbolicMulti]) -> list[SymbolicMulti]:
        combined: dict[str, SymbolicMulti] = {}
        for term in terms:
            k = term.signature
            if k in combined:
                combined[k] = term.reset_coef(combined[k].coef + term.coef)
            else:
                combined[k] = term
        return [v for v in combined.values() if not is_zero(v.coef)]

    def __add__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return SymbolicMultiSum([*self.terms, SymbolicMulti.pure(other)])

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([*self.terms, other])

        if isinstance(other, SymbolicMultiSum):
            return SymbolicMultiSum([*self.terms, *other.terms])
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return SymbolicMultiSum([*self.terms, SymbolicMulti.pure(-other)])

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([*self.terms, -1 * other])

        if isinstance(other, SymbolicMultiSum):
            terms = [*self.terms]
            terms.extend(-1 * term for term in other.terms)
            return SymbolicMultiSum(terms)
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return SymbolicMultiSum([SymbolicMulti.pure(other), *self.terms])

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([other, *self.terms])

        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            terms = [SymbolicMulti.pure(other)]
            terms.extend(-1 * term for term in self.terms)
            return SymbolicMultiSum(terms)

        if isinstance(other, SymbolicMulti):
            terms = [other]
            terms.extend(-1 * term for term in self.terms)
            return SymbolicMultiSum(terms)

        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            nother = as_numeric(other)
            if is_zero(nother):
                return 0
            if nother == 1:
                return self
            pv = self.pure_value()
            if pv is not None:
                return pv * nother
            return SymbolicMultiSum([x * as_real(nother) for x in self.terms])

        pv = self.pure_value()
        if pv is not None:
            return pv * other

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([x * other for x in self.terms])

        if isinstance(other, SymbolicMultiSum):
            ov = other.pure_value()
            if ov is not None:
                return self * ov

            # Combine like terms
            combined: dict[str, SymbolicMulti] = {}
            for term1 in self.terms:
                for term2 in other.terms:
                    prod = term1 * term2
                    k = prod.signature
                    if k in combined:
                        combined[k].coef += prod.coef
                    else:
                        combined[k] = prod
            terms = [v for v in combined.values() if not is_zero(v.coef)]
            return SymbolicMultiSum(terms)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            nother = as_numeric(other)
            if is_zero(nother):
                return SymbolicMultiSum.singleton(symbolic_zero)
            if nother == 1:
                return self
            pv = self.pure_value()
            if pv is not None:
                return SymbolicMultiSum.singleton(SymbolicMulti.pure(nother * pv))
            return SymbolicMultiSum([as_real(nother) * x for x in self.terms])

        pv = self.pure_value()
        if pv is not None:
            return other * pv

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([other * x for x in self.terms])

        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, int):
            if n == 0:
                return SymbolicMultiSum.singleton(symbolic_one)
            if n == 1:
                return self
            if n % 2 == 0:
                return (self * self) ** (n // 2)
            return self * (self * self) ** (n // 2)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if other == 1:
                return self
            d = as_real(other)
            return SymbolicMultiSum([term / d for term in self.terms])

        if isinstance(other, (SymbolicMulti, SymbolicMultiSum, SymbolicMultiRatio)):
            opv = other.pure_value()
            if opv is not None:
                if opv == 1:
                    return self
                return SymbolicMultiSum([term / opv for term in self.terms])
            return simplify(symbolic(self, other))

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if as_numeric(other) == 1:
                return symbolic(symbolic_one, self)
            d = as_real(other)
            return symbolic(SymbolicMulti.pure(d), self)

        if isinstance(other, (SymbolicMulti, SymbolicMultiRatio)):
            return simplify(symbolic(other, self))

        return NotImplemented


#
# Ratios of Sums of Multinomial Terms
#

class SymbolicMultiRatio(Symbolic):
    def __init__(self, numerator: SymbolicMultiSum, denominator: SymbolicMultiSum):
        terms = [numerator, denominator]
        # Simplify in the generic symbolic constructor so we can return
        # other types of values.
        self.terms = terms
        self.key = f'({numerator.key_of})/({denominator.key_of})'

        if numerator.is_pure() or numerator.is_single():
            num = str(numerator)
        else:
            num = f'({str(numerator)})'

        if denominator.is_pure() or denominator.is_single():
            den = str(denominator)
            if ' ' in den:
                den = f'({str(denominator)})'
        else:
            den = f'({str(denominator)})'

        self.as_str = f'{num}/{den}'

    @property
    def numerator(self):
        return self.terms[0]

    @property
    def denominator(self):
        return self.terms[1]

    @property
    def key_of(self):
        return self.key

    @property
    def sort_key(self):
        return (as_real(self.numerator.coef) / as_real(self.denominator.coef), self.key)

    def __str__(self) -> str:
        return self.as_str

    def __frplib_repr__(self) -> str:
        return str(self)

    def is_pure(self) -> bool:
        return self.numerator.is_pure() and self.denominator.is_pure()

    def pure_value(self):
        npv = self.numerator.pure_value()
        dpv = self.denominator.pure_value()

        if npv is not None and dpv is not None:
            return as_numeric(npv / dpv)
        return None

    def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> Symbolic | Numeric:
        spv = self.pure_value()
        if spv is not None:
            return spv if purify else SymbolicMultiSum.singleton(SymbolicMulti.pure(spv))

        num = self.numerator.substitute(mapping, purify)
        den = self.denominator.substitute(mapping, purify)

        # ATTN:Bug FIX 18 Jun2024: den was num here twice, use is_symbolic here instead?
        #      If num is purified, the symbolic call fails because num is a number
        if not isinstance(num, SymbolicMultiSum) and not isinstance(den, SymbolicMultiSum):
            return as_nice_numeric(as_real(num / den))
        if not is_symbolic(num):
            num = SymbolicMulti.pure(num)
        return simplify(symbolic(num, den))

    def __add__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                simplify(self)
            numer = self.numerator + as_real(other) * self.denominator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiRatio):
            if self.denominator.key_of == other.denominator.key_of:
                r = self.denominator.coef / other.denominator.coef
                return simplify(SymbolicMultiRatio(self.numerator + r * other.numerator, self.denominator))

            numer = self.numerator * other.denominator + self.denominator * other.numerator
            denom = self.denominator * other.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = self.numerator + other * self.denominator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            if not other:
                return self
            numer = self.numerator + SymbolicMultiSum.singleton(other) * self.denominator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                self
            numer = as_real(other) * self.denominator + self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = other * self.denominator + self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            if not other:
                return self
            numer = SymbolicMultiSum.singleton(other) * self.denominator + self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __sub__(self, other):
        return self + (-1 * other)

    def __rsub__(self, other):
        return other + (-1 * self)

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return symbolic_zero
            numer = self.numerator * as_real(other)
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiRatio):
            if self.numerator.key_of == other.denominator.key_of:
                return simplify(symbolic(other.numerator, self.denominator))

            if self.denominator.key_of == other.numerator.key_of:
                return simplify(symbolic(self.numerator, other.denominator))

            numer = self.numerator * other.numerator
            denom = self.denominator * other.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = self.numerator * other
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = self.numerator * SymbolicMultiSum.singleton(other)
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return symbolic_zero
            numer = as_real(other) * self.numerator
            denom = self.denominator
            return simplify(symbolic(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = other * self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = SymbolicMultiSum.singleton(other) * self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if other == 1:
                return self
            numer = self.numerator / as_numeric(other)
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiRatio):
            numer = self.numerator * other.denominator
            denom = self.denominator * other.numerator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = self.numerator
            denom = self.denominator * other
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = self.numerator
            denom = self.denominator * SymbolicMultiSum.singleton(other)
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            numer = SymbolicMultiSum.singleton(SymbolicMulti.pure(as_numeric(other)))
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer * denom, self.numerator))

        if isinstance(other, SymbolicMultiSum):
            numer = self.denominator * other
            denom = self.numerator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = self.denominator * SymbolicMultiSum.singleton(other)
            denom = self.numerator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, int):
            if n == 0:
                return 1
            if n == 1:
                return self
            return simplify(symbolic(self.numerator ** n, self.denominator ** n))
        return NotImplemented


#
# Simple Simplification Rules
#

def simplify(a: Symbolic | Numeric) -> Union[Symbolic, Numeric]:
    if isinstance(a, (int, Decimal)):
        return a

    apv = a.pure_value()
    if apv is not None:
        return as_numeric(apv)

    if isinstance(a, SymbolicMulti):
        return a

    if isinstance(a, SymbolicMultiSum):
        return a

    assert isinstance(a, SymbolicMultiRatio)
    npv = a.numerator.pure_value()
    dpv = a.denominator.pure_value()

    if npv is not None:
        if is_zero(npv):
            return 0
        return SymbolicMultiRatio(SymbolicMultiSum.singleton(symbolic_one), a.denominator / as_real(npv))
    if dpv is not None:
        return a.numerator / as_real(dpv)

    # We have a rational with neither term pure
    # First try to extract and cancel common terms
    common_num, adj_num = common_term(a.numerator)
    common_den, adj_den = common_term(a.denominator)
    simple_num, simple_den, changed = multis_divide(common_num, common_den)

    if adj_num.key_of == adj_den.key_of:
        ratio = as_real(adj_num.coef / adj_den.coef)
        return purify(simple_num) * ratio / purify(simple_den)   # type: ignore

    if not changed:
        return a

    return (purify(simple_num) / purify(simple_den)) * (adj_num / adj_den)   # type: ignore

    # Other simplification rules?

    # if a.numerator.key_of == a.denominator.key_of:
    #     return as_real(a.numerator.coef / a.denominator.coef)

    # return a


#
# Symbolic Constructors (use these not the class constructors)
#

def symbolic(numerator: Union[Symbolic, str], denominator: Union[Symbolic, Literal[1]] = 1) -> Union[Symbolic, Numeric]:
    if isinstance(numerator, str):
        numerator = symbol(numerator)

    npv = numerator.pure_value()
    dpv = denominator.pure_value() if denominator != 1 else 1
    if npv is not None and dpv is not None:
        return as_real(npv / dpv)
    elif npv is not None and denominator == 1:
        return npv
    elif npv is not None:
        numerator = symbolic_one
        denominator = denominator / npv
    elif denominator == 1:
        return simplify(numerator)

    assert isinstance(numerator, Symbolic)
    assert isinstance(denominator, Symbolic)

    if isinstance(numerator, SymbolicMulti):
        numerator = SymbolicMultiSum.singleton(numerator)

    if isinstance(denominator, SymbolicMulti):
        denominator = SymbolicMultiSum.singleton(denominator)

    num_rat = isinstance(numerator, SymbolicMultiRatio)
    den_rat = isinstance(denominator, SymbolicMultiRatio)

    if num_rat and den_rat and numerator.key_of == denominator.key_of:        # type: ignore
        return ((numerator.numerator.coef * denominator.denominator.coef) /   # type: ignore
                (numerator.denominator.coef * denominator.numerator.coef))    # type: ignore
    if num_rat:
        numerator = cast(SymbolicMultiSum, numerator)
        return simplify(numerator.__truediv__(denominator))
    elif den_rat:
        denominator = cast(SymbolicMultiSum, denominator)
        return simplify(denominator.__rtruediv__(numerator))

    numerator = cast(SymbolicMultiSum, numerator)
    denominator = cast(SymbolicMultiSum, denominator)

    if numerator.key_of == denominator.key_of:
        return as_real(numerator.coef / denominator.coef)

    return simplify(SymbolicMultiRatio(numerator, denominator))


#
# Symbol constructors
#

def gen_symbol() -> SymbolicMulti:
    "Generates a symbol with a unique name."
    var = next(gensym)
    return SymbolicMulti([var], [1])

def symbol(var: str) -> SymbolicMulti:
    "Generates a symbol with the given name, typically a single letter."
    return SymbolicMulti([var], [1])

def symbols(names: str) -> tuple[SymbolicMulti, ...]:
    """Generates symbols from the space-separated names in the given string.

    A pattern  <NAME>M ... <NAME>N with M <= N natural numbers, produces
    the symbols  NAME<M>, NAME<M+1>, ..., NAME<N>. For example,
    'h1 ... h4' gives symbols with names 'h1', 'h2', 'h3', and 'h4'.
    In this case, NAME should consist only of letters and '_''s.

    Returns a tuple of named symbols in the same order specified.

    """
    ELLIPSIS = '...'
    vars = names.strip().split()

    if vars[0] == ELLIPSIS:
        raise ConstructionError('Invalid ... at start of symbols argument.')

    n = len(vars)
    syms = []
    active = 0
    ind = 1
    while ind < n:
        if vars[ind] == ELLIPSIS:
            if ind + 1 >= n:
                raise ConstructionError('Invalid ... at end of symbols argument.')

            if vars[ind + 1] == ELLIPSIS:
                raise ConstructionError('Repeated ... in symbols argument.')

            v1 = re.match(r'([A-Za-z_]+)(0|[1-9][0-9]*)', vars[active])
            v2 = re.match(r'([A-Za-z_]+)(0|[1-9][0-9]*)', vars[ind + 1])

            if v1 is None or v2 is None:
                raise ConstructionError(f'Invalid pattern "{vars[active]} ... {vars[ind+1]}" in symbols argument.')

            if v1.group(1) != v2.group(1):
                raise ConstructionError(f'Mismatched names in "{vars[active]} ... {vars[ind+1]}" in symbols argument.')

            n1 = int(v1.group(2))
            n2 = int(v2.group(2))

            if n1 > n2:
                raise ConstructionError(f'Misordered indices in {n1} > {n2} in symbols argument.')

            syms.extend([symbol(f'{v1.group(1)}{j}') for j in range(n1, n2 + 1)])

            active = ind + 2

            if active >= n:
                break

            if vars[ind + 2] == ELLIPSIS:
                raise ConstructionError('Misplaced ... in symbols argument.')

            ind += 3
        else:
            syms.append(SymbolicMulti([vars[active]], [1]))
            active += 1
            ind += 1

    if active < n:
        syms.append(SymbolicMulti([vars[active]], [1]))

    # syms = [SymbolicMulti([var], [1]) for var in vars]
    return tuple(syms)

def is_symbolic(obj) -> TypeGuard[SymbolicMulti | SymbolicMultiSum | SymbolicMultiRatio]:
    return isinstance(obj, (SymbolicMulti, SymbolicMultiSum, SymbolicMultiRatio))


#
# Calculation Routines (may be subsumed by SymbolicFn)
#

def symbolic_sqrt(x: Symbolic) -> Symbolic:
    raise ConstructionError(f'Symbolic square root not yet implemented: {str(x)}')


#
# Info tags
#

setattr(symbol, '__info__', 'utilities::symbols')
setattr(gen_symbol, '__info__', 'utilities::symbols')



#
# Structure, Methods, Parsing for general (non-canonical) FRP Kinds
#
from __future__ import annotations

from collections       import namedtuple, defaultdict
from dataclasses       import dataclass
from enum              import Enum, auto
from fractions         import Fraction
from itertools         import accumulate
from typing            import Union
from typing_extensions import Any, TypeAlias

from parsy             import ParseError

from frplib.exceptions           import KindError
from frplib.numeric              import (Numeric, ScalarQ, as_nice_numeric, as_real,
                                         show_values, show_tuples)
from frplib.parsing.kind_strings import canonical_tree, kind_sexp, validate_kind
from frplib.parsing.parsy_adjust import parse_error_message
from frplib.quantity             import as_quantity, as_quant_vec, as_real_quantity
from frplib.symbolic             import Symbolic
from frplib.utils                import identity
from frplib.vec_tuples           import VecTuple, as_vec_tuple, vec_tuple

#
# Types
#


#
# Numeric Handling (move elsewhere)
#

# ATTN: Replace this with a suitable class
# ATTN: Need to move KindBranch to its own module to avoid circularity?

# Moved from kinds, needs refactoring

FRACTION_ROUNDING: int = 10**9

def rational_prob(x):  # ATTN: Add some methods to Numeric for this
    "Convert a probability to rational form. Some rounding of floats is done to get a nice value."
    if x < 1.0 / FRACTION_ROUNDING or 1 - x < x < 1.0 / FRACTION_ROUNDING:
        return Fraction(x)
    return Fraction(x).limit_denominator(FRACTION_ROUNDING)


#
# Representations of Weights and Values
#

# def as_weight(w: Union[ScalarQ, Symbolic] = as_numeric()):
#     if isinstance(w, Symbolic):
#         return w
#     return as_numeric(w)
#
# as_value = as_vec_tuple

as_weight = as_quantity
as_value = as_quant_vec


#
# Branch Representation in Canonical Form
#

_KindBranch = namedtuple('_KindBranch', ['vs', 'p'], defaults=(as_weight(1),))
# ATTN: Can type this out using the class form and NamedTuple superclass
# class _KindBranch(namedtuple):
#     vs: tuple,   # VecTuple[NumericD]
#     p: NumericD  # Probability


class KindBranch(_KindBranch):
    "A single branch in a Kind tree with a weight and a value"

    @classmethod
    def make(cls, vs, p):
        return KindBranch(vs=as_value(vs), p=as_weight(p))

    @property
    def value(self):
        "Extract the value from a KindBranch, returning a scalar if value tuple has dim 1."
        if len(self.vs) == 1:
            return self.vs[0]
        return self.vs

    def assoc(self, *, vs=None, p=None):
        "Returns new branch with values or prob updated by a function"
        have_v = vs is not None
        have_p = p  is not None
        if not have_v and not have_p:
            return self

        val = vs if have_v else self.vs
        prob = p if have_p else self.p
        return self.make(vs=val, p=prob)

    def update(self, *, vs=None, p=None):
        "Returns new branch with values or prob updated by a function"
        have_v = vs is not None
        have_p = p  is not None
        if not have_v and not have_p:
            return self

        val = vs(self.vs) if have_v else self.vs
        prob = p(self.p) if have_p else self.p
        return self.make(vs=val, p=prob)

    # (va -> vb) -> (pa -> pb) -> (KindBranch -> KindBranch)
    @staticmethod
    def bimap(vs=identity, p=identity):
        "Returns function that transforms values and probability of a kind branch"
        if vs is identity and p is identity:
            return identity

        def transform(branch):
            return KindBranch.make(vs=vs(branch.vs), p=p(branch.p))
        return transform

# ATTN: This version is not called; see kinds; remove?
def normalize_branches(canonical) -> list[KindBranch]:
    seen: dict[tuple, KindBranch] = {}
    total = as_quantity(sum(map(lambda b: b.p, canonical)), convert_numeric=as_real)
    for branch in canonical:
        if branch.vs in seen:
            seen[branch.vs] = KindBranch.make(vs=branch.vs, p=seen[branch.vs].p + branch.p / total)
        else:
            seen[branch.vs] = KindBranch.make(vs=branch.vs, p=branch.p / total)
    return sorted(seen.values(), key=lambda b: b.vs)


#
# Kind parsing from sexp string and tree format
#

def canonical_from_tree(ktree: list) -> list[KindBranch]:
    compact = canonical_tree(ktree, p=1,
                             weight_fn=as_real_quantity,
                             value_fn=as_real_quantity)
    canonical = [KindBranch.make(vs=subtree, p=weight) for weight, subtree in compact]
    canonical.sort(key=lambda x: x.vs)
    return canonical

def canonical_from_sexp(k_sexp_str: str) -> list[KindBranch]:
    try:
        ktree = kind_sexp.parse(k_sexp_str)
    except ParseError as e:
        raise KindError(parse_error_message(e))
    errors = validate_kind(ktree)
    if errors:
        raise KindError("".join(errors))

    return canonical_from_tree(ktree)


#
# Kind Output
#


#
# Kind Unfolding
#

def unfold_tree(canonical: list[KindBranch]) -> list | None:  # ATTN: give this a more specific type if needed
    """Unfolds a canonical kind tree to a sexp-tree, with info for printing.

    The format is similar to that produced by the sexp parsing functions, though
    the weighs are fractions and the values kept as is.

    Param:
      `canonical`: The canonical tree associated with a kind (._canonical).

    Returns an unfolded tree in sexp format, where each subtree has the
    form [node, *branches] and each branch is a pair [weight, leaf | subtree].
    However, if the canonical list is provided, returns None to signal that
    there is nothing left to do. This should short-circuit the analysis to
    use the standard methods for display canonical kinds.

    This is used as a primitive in a higher-level unfolding utility.

    """
    root = vec_tuple()
    if len(canonical) == 0 or len(canonical[0].vs) == 1:
        return None  # Nothing to do as we can already handle the canonical kind
        # return [root, *map(lambda b: [b.p, b.vs], canonical)]  # Unfolded tree but ...

    dim = len(canonical[0].vs)
    S = [[b.p, b.vs] for b in canonical]

    while len(S) > 0 and dim > 1:
        partition: dict[VecTuple, list] = defaultdict(list)
        weights: dict[VecTuple, Numeric] = defaultdict(as_nice_numeric)  # ATTN: handle symbolic?

        # Partition branches at this level into groups with common value prefix
        for branch in S:
            weight, subtree = branch
            prefix = subtree[0:-1] if isinstance(subtree, tuple) else subtree[0][0:-1]
            partition[prefix].append(branch)
            weights[prefix] += weight        # Sum will be weight to new parent node

        # Renormalize each partition element and insert new parent node from prefix
        T = []
        for prefix in partition:
            w = weights[prefix]
            nbranches = [[w_i / w, node] for w_i, node in partition[prefix]]
            subtree = [w, [prefix, *nbranches]]
            T.append(subtree)
            dim = len(prefix)  # They are all the same
        S = T

    unfolded = [root, *S]
    return unfolded

class Edge(Enum):
    FIRST = auto()
    MIDDLE = auto()
    OTHER = auto()
    LAST = auto()
    ROOT = auto()

@dataclass(frozen=True)
class Branch:
    x: int
    y: int
    edge: Edge
    level: int
    labels: tuple[str, str]

@dataclass(frozen=True)
class Segment:
    x: int
    y: int
    z: Any

Items: TypeAlias = list[Union[Branch, Segment]]
STree: TypeAlias = 'list[Union[str, STree]]'

# from frplib.kind_trees import unfold_tree
# from frplib.uscan import *
# from rich.pretty import pprint
# uu = uniform(1, 2, 3) * uniform(7, 8, 9)
# x = unfold_tree(uu._canonical)
# wd = [(0,3)]
# sep = [4, 2, 0]
# xl = unfolded_labels(x['unfolded'][1:], str(x['unfolded'][0]), 1, wd)
# s, _ = unfold_scan(xl, wd, sep)
# unfolded_str(s, wd)
#
# bit = choice(0, 1)
# bbb = bit ** 3
# z = unfold_tree(bbb._canonical)
# wd2 = [(0,3)]
# sep2 = [0, 4, 2, 0]
# zl = unfolded_labels(z['unfolded'][1:], str(z['unfolded'][0]), 1, wd2)
# s2, _ = unfold_scan(zl, wd2, sep2)
# unfolded_str(s2, wd2)


def unfolded_labels(unfolded_branches, root_str, level, widths) -> STree:
    w_strs = show_values(subtree[0] for subtree in unfolded_branches)
    v_strs = show_tuples([subtree[1] if isinstance(subtree[1], tuple) else subtree[1][0]
                          for subtree in unfolded_branches], scalarize=False)
    w_max = max(len(w) for w in w_strs) + 2  # for spaces
    v_max = max(len(v) for v in v_strs) + 2  # for spaces
    w_strs = ['{0:-<{wd}}'.format(' ' + w + ' ', wd=w_max) for w in w_strs]
    v_strs = ['{0:<{wd}}'.format(' ' + v + ' ', wd=v_max)  for v in v_strs]

    if len(widths) <= level:
        widths.append((w_max, v_max))
        assert len(widths) >= level  # must be length >= level - 1 to get to append
    else:
        widths[level] = tuple(map(max, zip(widths[level], (w_max, v_max))))

    out: STree = []
    for index, branch in enumerate(unfolded_branches):
        if isinstance(branch[1], tuple):
            out.append([w_strs[index], v_strs[index]])
        else:
            out.append([w_strs[index], unfolded_labels(branch[1][1:], v_strs[index], level + 1, widths)])
    return [root_str, *out]

def after_dashes(weight_width: int) -> int:
    if weight_width < 6:
        return 4
    elif weight_width < 10:
        return 6
    else:
        return 8

def unfold_scan(unfolded, widths_by_level: list[tuple[int, int]], sep: list[int]) -> tuple[Items, list[int]]:
    "Converts an unfolded tree into scan ordered branches for display."

    def level_width(level, leaf=False):
        if level == 0:
            return 4
        weight_width, value_width = widths_by_level[level]
        extra = 1 + (1 - leaf)  # 1 for segment line, 1 for next node connection
        dashes = after_dashes(weight_width)  # number dashes after weight
        return extra + (2 * dashes + 1) + weight_width + value_width

    num_levels = len(widths_by_level)
    level_widths = [level_width(level, level == num_levels - 1)
                    for level in range(num_levels)]
    x_by_level = list(accumulate(level_widths, initial=0))

    def scan(index: tuple[int, int, int], tree: STree, acc: Items) -> tuple[Branch, int, int]:
        # index = (m, ell) the mth branch at level ell
        # tree = [w, leaf | subtree]:  tree contains *labels* as data, computed together by subtree!!
        # acc = list of Items

        m, size, level = index
        x = x_by_level[level]
        weight, more = tree
        assert isinstance(weight, str)

        if level == 0:
            edge = Edge.ROOT
        elif m == 0 and size > 1:
            edge = Edge.FIRST
        elif size % 2 != 0 and m == size // 2:
            edge = Edge.MIDDLE
        elif m == size - 1:
            edge = Edge.LAST
        else:
            edge = Edge.OTHER

        # Leaf
        if isinstance(more, str):  # <=> level == env.dim
            if len(acc) == 0:  # First/top leaf node
                return (Branch(x, 0, edge, level, (weight, more)), 0, 0)
            base = acc[-1]
            yb = base.y
            return (Branch(x, yb + 1, edge, level, (weight, more)), yb + 1, yb + 1)

        # Non-Leaf
        node, *subtree = more
        assert isinstance(node, str)
        assert len(subtree) > 0

        size_prime = len(subtree)
        no_middle = size_prime % 2 == 0
        half_height = size_prime // 2
        x_prime = x_by_level[level + 1]
        y_min = int(10 ** 10)  # Sentinels
        y_max = -1

        for k in range(half_height):
            b = subtree[k]
            assert not isinstance(b, str)
            b_prime, y_max, y_min_k = scan((k, size_prime, level + 1), b, acc)
            y_min = min(y_min, y_min_k)
            acc.append(b_prime)
            # k always < size_prime - 1 here
            if no_middle and k == half_height - 1:
                yb = y_max
                for i in range(sep[level + 1] // 2):
                    yb += 1
                    acc.append(Segment(x_prime, yb, 'D'))
            else:
                yb = y_max
                for i in range(sep[level + 1]):
                    yb += 1
                    acc.append(Segment(x_prime, yb, 'D'))

        # k == half_height
        if no_middle:
            yb += 1
            me = Branch(x, yb, edge, level, (weight, node))
            y_seg = me.y + 1
            if m > 0:
                for y in range(y_min, me.y):
                    acc.append(Segment(x, y, 'B'))
            acc.append(Segment(x_prime, yb, 'A'))  # --| split
            for i in range(sep[level + 1] // 2, sep[level + 1]):
                yb += 1
                acc.append(Segment(x_prime, yb, 'D'))

        for k in range(half_height, size_prime):
            b = subtree[k]
            assert not isinstance(b, str)
            b_prime, y_max, y_min_k = scan((k, size_prime, level + 1), b, acc)
            yb = b_prime.y
            y_min = min(y_min, y_min_k)

            if k == half_height and not no_middle:
                me = Branch(x, yb, edge, level, (weight, node))
                if m > 0:
                    for y in range(y_min, me.y):
                        acc.append(Segment(x, y, 'B'))
                y_seg = me.y + 1

            if m < size - 1:
                # ATTN! Overlap here as y_max grows
                for y in range(y_seg, y_max + 1):
                    acc.append(Segment(x, y, 'C'))
                y_seg = y_max + 1
            acc.append(b_prime)

            if k < size_prime - 1:
                yb = y_max
                for i in range(sep[level + 1]):
                    yb += 1
                    acc.append(Segment(x_prime, yb, 'D'))

        # for k, b in enumerate(subtree):
        #     assert not isinstance(b, str)
        #     if k == half_height and no_middle:
        #         # yb is height of last branch above the middle
        #         # ATTN: yb + 1 ?  yb for the first helps some things and hurts others
        #         me = Branch(x, yb + 1, edge, level, (weight, node))
        #         acc.append(Segment(x_prime, yb + 1, 'A'))  # --| split
        #
        #     b_prime, y_max, y_min_k = scan((k, size_prime, level + 1), b, acc)
        #     yb = b_prime.y
        #     y_min = min(y_min, y_min_k)
        #
        #     if k < half_height:
        #         # if m > 0:
        #         #     acc.append(Segment(x, yb, 'B'))
        #         acc.append(b_prime)
        #     if k == half_height and not no_middle:
        #         me = Branch(x, yb, edge, level, (weight, node))
        #     if k == half_height:
        #         if m > 0:
        #             for y in range(y_min, me.y):
        #                 acc.append(Segment(x, y, 'B'))
        #     if k >= half_height:
        #         if m < size - 1:
        #             # ATTN! Overlap here as y_max grows
        #             for y in range(me.y + 1, y_max + 1):
        #                 acc.append(Segment(x, y, 'C'))
        #             # acc.append(Segment(x, yb, 'C'))
        #         acc.append(b_prime)
        #
        #     if k < size_prime - 1:
        #         yb = y_max
        #         for i in range(sep[level + 1]):
        #             yb += 1
        #             acc.append(Segment(x_prime, yb, 'D'))
        return (me, max(yb, y_max), y_min)

    if len(unfolded) == 1:
        return ([Branch(4, 0, Edge.ROOT, 0, ('', unfolded[0]))], level_widths)

    items: Items = []
    root, _, _ = scan((0, 1, 0), ['', unfolded], items)
    items.append(root)
    items.sort(key=lambda v: (v.y, v.x))

    return (items, level_widths)

def unfolded_str(scanned: Items, widths_by_level: list[tuple[int, int]]) -> str:
    # Temporarily interpret items here; move to another function
    y_last = 0
    x_last = 0
    num_levels = len(widths_by_level)

    out: list[str] = []
    for item in scanned:
        x, y = (item.x, item.y)
        # Fill in missing spaces
        if y_last == y:
            out.append(' ' * (x - x_last))
        else:
            out.append('\n' * (y - y_last))
            out.append(' ' * x)

        # Print appropriate piece for this item:
        #   Branch(x, y, First,  level labels) =>    ,----- w ---- <...>   -
        #   Branch(x, y, Middle, level labels) =>    +----- w ---- <...>   -
        #   Branch(x, y, Last,   level labels) =>    `----- w ---- <...>   -
        #   Branch(x, y, Other,  level labels) =>    |----- w ---- <...>   -
        #   Segment(x, y)                      =>    |
        # Skip the final joining piece for a lief

        if isinstance(item, Segment):
            token = '|'
        else:
            edge = item.edge
            level = item.level
            connect = '-' if level < num_levels - 1 else ''
            wlabel, vlabel = item.labels
            after_dash_s = '-' * after_dashes(len(wlabel))
            before_dash_s = after_dash_s + '-'
            joins = {Edge.FIRST: ',', Edge.MIDDLE: '+', Edge.LAST: '`', Edge.OTHER: '|'}

            if edge == Edge.ROOT:
                token = '<> {0}'.format(connect)
            else:
                token = '{seg}{before}{weight:-<{wwd}}{after}{value:<{vwd}}{con}'.format(
                    seg=joins[edge],
                    before=before_dash_s, weight=wlabel, after=after_dash_s,
                    value=vlabel, con=connect,
                    wwd=widths_by_level[level][0],
                    vwd=widths_by_level[level][1]
                )
        out.append(token)
        x_last = x + len(token)
        y_last = y
    return "".join(out)



# Assistant Assistance Example Chapter 0 Section 8

from typing              import cast

from frplib.expectations import E
from frplib.kinds        import conditional_kind, constant, permutations_of, uniform
from frplib.statistics   import condition, statistic, ArgMax, ArgMin, Proj
from frplib.utils        import irange
from frplib.vec_tuples   import as_scalar_strict, as_vec_tuple, vec_tuple

# Solution to the problem

def is_success(k):
    "Statistic factory that checks if After-k succeeds given <b,s>."

    @condition
    def succeeded(b, s):
        return b > k and s <= k

    return succeeded

def assistant(n):
    "Returns success probabilities of After-k strategies with `n` applicants."
    assert n >= 1, "at least one applicant is required"

    best = uniform(irange(n))

    @conditional_kind(codim=1)
    def pre_best_position(m):
        if m <= 1:
            return constant(0)
        return uniform(irange(1, m - 1))

    best_pre_best = best >> pre_best_position

    success_probs = [0] * n
    for k in range(n):
        success_probs[k] = E(best_pre_best ^ is_success(k)).raw

    return as_vec_tuple(success_probs)

def best_k(n):
    "Finds the best k and probability for the After-k strategy with n applicants"
    success_probs = assistant(n)
    k_star = as_scalar_strict(ArgMax(success_probs))
    return vec_tuple(k_star, success_probs[k_star])

# Direct approach: Checks that best position and the pre-best position are uniform

@statistic
def best_pre_best_of(v):
    b = cast(int, as_scalar_strict(ArgMin(v)))
    if b == 0:
        return (b + 1, 0)
    return (b + 1, as_scalar_strict(ArgMin(v[:b])) + 1)

def check_best_position(n=7):
    permutations_of(irange(1, 7)) ^ best_pre_best_of ^ Proj[1]

def check_pre_best_position(b, n=7):
    Proj[2] @ (permutations_of(irange(1, 7)) ^ best_pre_best_of) | (Proj[1] == b)



from __future__ import annotations

from typing            import (
    Protocol,
    runtime_checkable
)

from collections.abc   import Collection


@runtime_checkable
class Projection(Protocol):
    @property
    def subspace(self) -> Collection[int]:
        ...

## ATTN! add *a, **kw to this signature for utility purposes
@runtime_checkable
class Renderable(Protocol):
    def __frplib_repr__(self):
        ...

@runtime_checkable
class Transformable(Protocol):
    def transform(self, f_mapping):
        ...

@runtime_checkable
class SupportsExpectation(Protocol):
    def expectation(self):
        ...

@runtime_checkable
class SupportsApproxExpectation(Protocol):
    def approximate_expectation(self, tolerance):
        ...

@runtime_checkable
class SupportsForcedExpectation(Protocol):
    def forced_expectation(self):
        ...



# output.py - tools for managing terminal output

from __future__ import annotations

from dataclasses       import dataclass
from decimal           import Decimal
from typing            import Literal
from typing_extensions import Any

from rich              import box
from rich.panel        import Panel

from frplib.env        import environment
from frplib.numeric    import as_nice_numeric

#
# Rendered Output
#

def in_panel(
        s: str,
        box=box.SQUARE,
        title: str | None = None,
        title_align: Literal['left', 'center', 'right'] = 'center',
        subtitle: str | None = None,
        subtitle_align: Literal['left', 'center', 'right'] = 'center',
) -> str | Panel:
    if environment.ascii_only:
        return s
    return Panel(
        s,
        expand=False,
        box=box,
        title=title,
        title_align=title_align,
        subtitle=subtitle,
        subtitle_align=subtitle_align,
    )


#
# Wrapped Quantities Providing Rich String Representations
#

@dataclass(frozen=True)
class RichQuantity:
    this: Any

    def __str__(self) -> str:
        return str(self.this)

    def __repr__(self) -> str:
        return repr(self.this)

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

@dataclass(frozen=True)
class TitledRichQuantity:
    this: Any
    title: str = ''

    def __str__(self) -> str:
        return self.title + str(self.this)

    def __repr__(self) -> str:
        return repr(self.this)

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

@dataclass(frozen=True)
class RichFacade:
    this: Any
    facade: str = ''

    def __str__(self) -> str:
        if self.facade:
            return self.facade
        else:
            return str(self.this)

    def __repr__(self) -> str:
        return repr(self.this)

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

@dataclass(frozen=True)
class TitledRichFacade:
    this: Any
    facade: str = ''
    title: str = ''

    def __str__(self) -> str:
        if self.facade:
            return self.title + self.facade
        else:
            return self.title + str(self.this)

    def __repr__(self) -> str:
        return repr(self.this)

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

class RichString(str):
    def __frplib_repr__(self):
        return str(self)

class RichReal(Decimal):
    def __frplib_repr__(self):
        return str(as_nice_numeric(self))



from __future__ import annotations

from itertools import chain

from frplib.examples.roulette import roulette, RED_SQUARES

from frplib.kinds      import Kind, kind
from frplib.utils      import irange
from frplib.vec_tuples import vec_tuple

def test_roulette_basics():
    assert Kind.equal(roulette.kind, kind(roulette()))

def test_roulette_plays():
    lost = vec_tuple(-1)
    won = vec_tuple(1)    # scale as needed

    assert all( roulette.even(p) == won for p in irange(2, 36, step=2) )
    assert all( roulette.even(p) == lost for p in irange(-1, 36, step=2) )
    assert roulette.even(0) == lost
    assert all( roulette.odd(p) == won for p in irange(1, 36, step=2) )
    assert all( roulette.odd(p) == lost for p in irange(2, 36, step=2) )
    assert roulette.odd(0) == lost
    assert roulette.odd(-1) == lost

    assert all( roulette.red(p) == won for p in RED_SQUARES )
    assert all( roulette.red(p) == lost for p in irange(-1, 36) if p not in RED_SQUARES )
    assert all( roulette.black(p) == won for p in irange(1, 36) if p not in RED_SQUARES )
    assert all( roulette.black(p) == lost for p in RED_SQUARES )
    assert roulette.black(0) == lost
    assert roulette.black(-1) == lost

    assert all( roulette.first18(p) == won for p in irange(1, 18) )
    assert all( roulette.second18(p) == won for p in irange(19, 36) )
    assert all( roulette.first18(p) == lost for p in irange(-1, 36) if p < 1 or p > 18)
    assert all( roulette.second18(p) == lost for p in irange(-1, 18) )

    assert all( roulette.dozen(1) == 2 * won for p in irange(1, 12) )
    assert all( roulette.dozen(2) == 2 * won for p in irange(13, 24) )
    assert all( roulette.dozen(3) == 2 * won for p in irange(25, 36) )
    assert all( roulette.dozen('first') == 2 * won for p in irange(1, 12) )
    assert all( roulette.dozen('second') == 2 * won for p in irange(13, 24) )
    assert all( roulette.dozen('third') == 2 * won for p in irange(25, 36) )
    assert all( roulette.dozen(1) == lost for p in chain([-1, 0], irange(13, 36)) )
    assert all( roulette.dozen(2) == lost for p in chain(irange(-1, 12), irange(25, 36)) )
    assert all( roulette.dozen(3) == lost for p in irange(-1, 24) )

    assert all( roulette.column(1) == 2 * won for p in irange(1, 36) if p % 3 == 1 )
    assert all( roulette.column(2) == 2 * won for p in irange(1, 36) if p % 3 == 2 )
    assert all( roulette.column(3) == 2 * won for p in irange(1, 36) if p % 3 == 0 )
    assert all( roulette.column(1) == lost    for p in irange(1, 36) if p % 3 != 1 )
    assert all( roulette.column(2) == lost    for p in irange(1, 36) if p % 3 != 2 )
    assert all( roulette.column(3) == lost    for p in irange(1, 36) if p % 3 != 0 )
    assert all( roulette.column(i)(0) == lost  for i in irange(1, 3) )
    assert all( roulette.column(i)(-1) == lost for i in irange(1, 3) )

    assert all( roulette.top_line == 5 * won for p in irange(-1, 3) )
    assert all( roulette.top_line == lost for p in irange(4, 36) )

    assert all( roulette.six_line(p)(-1) == lost for p in irange(1, 36) )
    assert all( roulette.six_line(p)(0) == lost for p in irange(1, 36) )

    # ATTN: finish six_line, corner, street, and split

    assert all( roulette.straight(p)(p) == 35 * won for p in irange(-1, 36) )
    assert all( roulette.straight(p)(q) == lost for p in irange(-1, 36) for q in irange(-1, 36) if p != q )



#!/usr/bin/env python
"""
A stub for the playground app, which will actually be a subcommand.
"""

from __future__ import annotations

from pathlib import Path

from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts      import print_formatted_text as print
from ptpython.prompt_style         import PromptStyle

from frplib.env                    import environment
from frplib.repls.playground_repl  import PlaygroundRepl
from frplib.repls.ptembed          import embed


def style_prompt_bright(text):
    return f'<steelblue>{text}</steelblue>'

def style_prompt_dark(text):
    return f'<style fg="#b97d4b">{text}</style>'

def configure(repl):
    # Probably, the best is to add a new PromptStyle to `all_prompt_styles` and
    # activate it. This way, the other styles are still selectable from the
    # menu.
    class CustomPrompt(PromptStyle):
        def in_prompt(self):
            style_prompt = style_prompt_dark if environment.dark_mode else style_prompt_bright
            return HTML(style_prompt("playground&gt; "))

        def in2_prompt(self, width):
            style_prompt = style_prompt_dark if environment.dark_mode else style_prompt_bright
            return HTML(style_prompt("...&gt; ".ljust(width)))

        def out_prompt(self):
            return []

    repl.all_prompt_styles["playground"] = CustomPrompt()
    repl.prompt_style = "playground"
    repl.show_signature = False
    repl.show_docstring = True
    repl.enable_syntax_highlighting = True
    repl.highlight_matching_parenthesis = True

    if environment.dark_mode:
        repl.use_code_colorscheme("github-dark")
    else:
        repl.use_code_colorscheme("default")

    # Title in status bar
    repl.title = HTML('<style fg="#0099ff"><b>FRP Playground</b></style> ')

def main():
    try:
        embed(
            globals(),
            locals(),
            title='FRP Playground',
            configure=configure,
            make_repl=PlaygroundRepl,
            history_filename=str(Path.home() / ".frp-playground-history")
        )
    except SystemExit:
        pass
    print('Playground finished')


if __name__ == "__main__":
    main()



from __future__ import annotations

from decimal import Decimal

from frplib.kinds      import sequence_of_values, Flatten
from frplib.numeric    import as_numeric
from frplib.symbolic   import symbol, is_symbolic
from frplib.utils      import every, lmap
from frplib.vec_tuples import as_vec_tuple, vec_tuple


#
# Helpers
#

def numeric(x):
    return x if x is Ellipsis else as_numeric(x)

def test_simple_seq():
    "Simple sequence patterns."
    assert (sequence_of_values(1, 2, ..., 8) ==
            [1, 2, 3, 4, 5, 6, 7, 8])
    assert (sequence_of_values((1, 2), (3, 4), (5, 6), (7, 8)) ==
            [1, 2, 3, 4, 5, 6, 7, 8])
    assert (sequence_of_values(1, 2, ..., 10, 12, ..., 22) ==
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 14, 16, 18, 20, 22])
    assert (sequence_of_values((1, 2), (3, 4), (5, 6), (7, 8),
                               pre_transform=as_vec_tuple) ==
            [vec_tuple(1, 2), vec_tuple(3, 4), vec_tuple(5, 6), vec_tuple(7, 8)])
    assert (sequence_of_values((1, 2), (3, 4), (5, 6), (7, 8),
                               flatten=Flatten.EVERYTHING, pre_transform=as_vec_tuple) ==
            [1, 2, 3, 4, 5, 6, 7, 8])
    assert (sequence_of_values('1.2', '2.4', ..., '10',
                               pre_transform=numeric) ==
            [Decimal('1.2'), Decimal('2.4'), Decimal('3.6'), Decimal('4.8'),
             Decimal('6.0'), Decimal('7.2'), Decimal('8.4'), 10])

def test_symbol_seq():
    a = symbol('a')
    b = symbol('b')
    assert every(is_symbolic, sequence_of_values(a, b, 2 * a))
    assert lmap(str, sequence_of_values(a, b, 2 * a)) == ['a', 'b', '2 a']
